@@ -1,6 +1,6 @@
 /**
  * @file libimobiledevice/mobilebackup.h
- * @brief MobileBackup Implementation
+ * @brief Backup and restore of all device data.
  * \internal
  *
  * Copyright (c) 2009 Martin Szulecki All Rights Reserved.
@@ -28,6 +28,9 @@ extern "C" {
 #endif
 
 #include <libimobiledevice/libimobiledevice.h>
+#include <libimobiledevice/lockdown.h>
+
+#define MOBILEBACKUP_SERVICE_NAME "com.apple.mobilebackup"
 
 /** @name Error Codes */
 /*@{*/
@@ -47,12 +50,24 @@ typedef int16_t mobilebackup_error_t;
 typedef struct mobilebackup_client_private mobilebackup_client_private;
 typedef mobilebackup_client_private *mobilebackup_client_t; /**< The client handle. */
 
-mobilebackup_error_t mobilebackup_client_new(idevice_t device, uint16_t port, mobilebackup_client_t * client);
+typedef enum {
+	MB_RESTORE_NOTIFY_SPRINGBOARD = 1 << 0,
+	MB_RESTORE_PRESERVE_SETTINGS = 1 << 1,
+	MB_RESTORE_PRESERVE_CAMERA_ROLL = 1 << 2
+} mobilebackup_flags_t;
+
+mobilebackup_error_t mobilebackup_client_new(idevice_t device, lockdownd_service_descriptor_t service, mobilebackup_client_t * client);
+mobilebackup_error_t mobilebackup_client_start_service(idevice_t device, mobilebackup_client_t* client, const char* label);
 mobilebackup_error_t mobilebackup_client_free(mobilebackup_client_t client);
+
 mobilebackup_error_t mobilebackup_receive(mobilebackup_client_t client, plist_t *plist);
 mobilebackup_error_t mobilebackup_send(mobilebackup_client_t client, plist_t plist);
 mobilebackup_error_t mobilebackup_request_backup(mobilebackup_client_t client, plist_t backup_manifest, const char *base_path, const char *proto_version);
 mobilebackup_error_t mobilebackup_send_backup_file_received(mobilebackup_client_t client);
+mobilebackup_error_t mobilebackup_request_restore(mobilebackup_client_t client, plist_t backup_manifest, mobilebackup_flags_t flags, const char *proto_version);
+mobilebackup_error_t mobilebackup_receive_restore_file_received(mobilebackup_client_t client, plist_t *result);
+mobilebackup_error_t mobilebackup_receive_restore_application_received(mobilebackup_client_t client, plist_t *result);
+mobilebackup_error_t mobilebackup_send_restore_complete(mobilebackup_client_t client);
 mobilebackup_error_t mobilebackup_send_error(mobilebackup_client_t client, const char *reason);
 
 #ifdef __cplusplus
