@@ -2,6 +2,7 @@
  * installation_proxy.c
  * com.apple.mobile.installation_proxy service implementation.
  *
+ * Copyright (c) 2013 Martin Szulecki All Rights Reserved.
  * Copyright (c) 2009 Nikias Bassen, All Rights Reserved.
  *
  * This library is free software; you can redistribute it and/or
@@ -26,12 +27,13 @@
 
 #include "installation_proxy.h"
 #include "property_list_service.h"
-#include "debug.h"
+#include "common/debug.h"
 
 struct instproxy_status_data {
 	instproxy_client_t client;
 	instproxy_status_cb_t cbfunc;
 	char *operation;
+	void *user_data;
 };
 
 /**
@@ -42,7 +44,7 @@ struct instproxy_status_data {
 static void instproxy_lock(instproxy_client_t client)
 {
 	debug_info("InstallationProxy: Locked");
-	g_mutex_lock(client->mutex);
+	mutex_lock(&client->mutex);
 }
 
 /**
@@ -53,7 +55,7 @@ static void instproxy_lock(instproxy_client_t client)
 static void instproxy_unlock(instproxy_client_t client)
 {
 	debug_info("InstallationProxy: Unlocked");
-	g_mutex_unlock(client->mutex);
+	mutex_unlock(&client->mutex);
 }
 
 /**
@@ -76,6 +78,8 @@ static instproxy_error_t instproxy_error(property_list_service_error_t err)
 			return INSTPROXY_E_PLIST_ERROR;
 		case PROPERTY_LIST_SERVICE_E_MUX_ERROR:
 			return INSTPROXY_E_CONN_FAILED;
+		case PROPERTY_LIST_SERVICE_E_RECEIVE_TIMEOUT:
+			return INSTPROXY_E_RECEIVE_TIMEOUT;
 		default:
 			break;
 	}
@@ -86,36 +90,50 @@ static instproxy_error_t instproxy_error(property_list_service_error_t err)
  * Connects to the installation_proxy service on the specified device.
  *
  * @param device The device to connect to
- * @param port Destination port (usually given by lockdownd_start_service).
+ * @param service The service descriptor returned by lockdownd_start_service.
  * @param client Pointer that will be set to a newly allocated
  *     instproxy_client_t upon successful return.
  *
  * @return INSTPROXY_E_SUCCESS on success, or an INSTPROXY_E_* error value
  *     when an error occured.
  */
-instproxy_error_t instproxy_client_new(idevice_t device, uint16_t port, instproxy_client_t *client)
+instproxy_error_t instproxy_client_new(idevice_t device, lockdownd_service_descriptor_t service, instproxy_client_t *client)
 {
-	/* makes sure thread environment is available */
-	if (!g_thread_supported())
-		g_thread_init(NULL);
-
-	if (!device)
-		return INSTPROXY_E_INVALID_ARG;
-
 	property_list_service_client_t plistclient = NULL;
-	if (property_list_service_client_new(device, port, &plistclient) != PROPERTY_LIST_SERVICE_E_SUCCESS) {
-		return INSTPROXY_E_CONN_FAILED;
+	instproxy_error_t err = instproxy_error(property_list_service_client_new(device, service, &plistclient));
+	if (err != INSTPROXY_E_SUCCESS) {
+		return err;
 	}
 
 	instproxy_client_t client_loc = (instproxy_client_t) malloc(sizeof(struct instproxy_client_private));
 	client_loc->parent = plistclient;
-	client_loc->mutex = g_mutex_new();
-	client_loc->status_updater = NULL;
+	mutex_init(&client_loc->mutex);
+	client_loc->status_updater = (thread_t)NULL;
 
 	*client = client_loc;
 	return INSTPROXY_E_SUCCESS;
 }
 
+/**
+ * Starts a new installation_proxy service on the specified device and connects to it.
+ *
+ * @param device The device to connect to.
+ * @param client Pointer that will point to a newly allocated
+ *     instproxy_client_t upon successful return. Must be freed using
+ *     instproxy_client_free() after use.
+ * @param label The label to use for communication. Usually the program name.
+ *  Pass NULL to disable sending the label in requests to lockdownd.
+ *
+ * @return INSTPROXY_E_SUCCESS on success, or an INSTPROXY_E_* error
+ *     code otherwise.
+ */
+instproxy_error_t instproxy_client_start_service(idevice_t device, instproxy_client_t * client, const char* label)
+{
+	instproxy_error_t err = INSTPROXY_E_UNKNOWN_ERROR;
+	service_client_factory_start_service(device, INSTPROXY_SERVICE_NAME, (void**)client, label, SERVICE_CONSTRUCTOR(instproxy_client_new), &err);
+	return err;
+}
+
 /**
  * Disconnects an installation_proxy client from the device and frees up the
  * installation_proxy client data.
@@ -134,11 +152,9 @@ instproxy_error_t instproxy_client_free(instproxy_client_t client)
 	client->parent = NULL;
 	if (client->status_updater) {
 		debug_info("joining status_updater");
-		g_thread_join(client->status_updater);
-	}
-	if (client->mutex) {
-		g_mutex_free(client->mutex);
+		thread_join(client->status_updater);
 	}
+	mutex_destroy(&client->mutex);
 	free(client);
 
 	return INSTPROXY_E_SUCCESS;
@@ -164,14 +180,14 @@ static instproxy_error_t instproxy_send_command(instproxy_client_t client, const
 
 	plist_t dict = plist_new_dict();
 	if (appid) {
-		plist_dict_insert_item(dict, "ApplicationIdentifier", plist_new_string(appid));
+		plist_dict_set_item(dict, "ApplicationIdentifier", plist_new_string(appid));
 	}
 	if (client_options && (plist_dict_get_size(client_options) > 0)) {
-		plist_dict_insert_item(dict, "ClientOptions", plist_copy(client_options));
+		plist_dict_set_item(dict, "ClientOptions", plist_copy(client_options));
 	}
-	plist_dict_insert_item(dict, "Command", plist_new_string(command));
+	plist_dict_set_item(dict, "Command", plist_new_string(command));
 	if (package_path) {
-		plist_dict_insert_item(dict, "PackagePath", plist_new_string(package_path));
+		plist_dict_set_item(dict, "PackagePath", plist_new_string(package_path));
 	}
 
 	instproxy_error_t err = instproxy_error(property_list_service_send_xml_plist(client->parent, dict));
@@ -215,7 +231,7 @@ instproxy_error_t instproxy_browse(instproxy_client_t client, plist_t client_opt
 		browsing = 0;
 		dict = NULL;
 		res = instproxy_error(property_list_service_receive_plist(client->parent, &dict));
-		if (res != INSTPROXY_E_SUCCESS) {
+		if (res != INSTPROXY_E_SUCCESS && res != INSTPROXY_E_RECEIVE_TIMEOUT) {
 			break;
 		}
 		if (dict) {
@@ -252,6 +268,8 @@ instproxy_error_t instproxy_browse(instproxy_client_t client, plist_t client_opt
 
 	if (res == INSTPROXY_E_SUCCESS) {
 		*result = apps_array;
+	} else {
+		plist_free(apps_array);
 	}
 
 leave_unlock:
@@ -270,8 +288,9 @@ leave_unlock:
  * @param status_cb Pointer to a callback function or NULL
  * @param operation Operation name. Will be passed to the callback function
  *        in async mode or shown in debug messages in sync mode. 
+ * @param user_data Callback data passed to status_cb.
  */
-static instproxy_error_t instproxy_perform_operation(instproxy_client_t client, instproxy_status_cb_t status_cb, const char *operation)
+static instproxy_error_t instproxy_perform_operation(instproxy_client_t client, instproxy_status_cb_t status_cb, const char *operation, void *user_data)
 {
 	instproxy_error_t res = INSTPROXY_E_UNKNOWN_ERROR;
 	int ok = 1;
@@ -279,16 +298,16 @@ static instproxy_error_t instproxy_perform_operation(instproxy_client_t client,
 
 	do {
 		instproxy_lock(client);
-		res = instproxy_error(property_list_service_receive_plist_with_timeout(client->parent, &dict, 30000));
+		res = instproxy_error(property_list_service_receive_plist_with_timeout(client->parent, &dict, 1000));
 		instproxy_unlock(client);
-		if (res != INSTPROXY_E_SUCCESS) {
+		if (res != INSTPROXY_E_SUCCESS && res != INSTPROXY_E_RECEIVE_TIMEOUT) {
 			debug_info("could not receive plist, error %d", res);
 			break;
 		}
 		if (dict) {
 			/* invoke callback function */
 			if (status_cb) {
-				status_cb(operation, dict);
+				status_cb(operation, dict, user_data);
 			}
 			/* check for 'Error', so we can abort cleanly */
 			plist_t err = plist_dict_get_item(dict, "Error");
@@ -347,20 +366,20 @@ static instproxy_error_t instproxy_perform_operation(instproxy_client_t client,
  *
  * @return Always NULL.
  */
-static gpointer instproxy_status_updater(gpointer arg)
+static void* instproxy_status_updater(void* arg)
 {	
 	struct instproxy_status_data *data = (struct instproxy_status_data*)arg;
 
 	/* run until the operation is complete or an error occurs */
-	(void)instproxy_perform_operation(data->client, data->cbfunc, data->operation);
+	(void)instproxy_perform_operation(data->client, data->cbfunc, data->operation, data->user_data);
 
 	/* cleanup */
 	instproxy_lock(data->client);
 	debug_info("done, cleaning up.");
 	if (data->operation) {
-	    free(data->operation);
+		free(data->operation);
 	}
-	data->client->status_updater = NULL;
+	data->client->status_updater = (thread_t)NULL;
 	instproxy_unlock(data->client);
 	free(data);
 
@@ -377,12 +396,13 @@ static gpointer instproxy_status_updater(gpointer arg)
  * @param status_cb Pointer to a callback function or NULL
  * @param operation Operation name. Will be passed to the callback function
  *        in async mode or shown in debug messages in sync mode.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS when the thread was created (async mode), or
  *         when the operation completed successfully (sync).
  *         An INSTPROXY_E_* error value is returned if an error occured.
  */
-static instproxy_error_t instproxy_create_status_updater(instproxy_client_t client, instproxy_status_cb_t status_cb, const char *operation)
+static instproxy_error_t instproxy_create_status_updater(instproxy_client_t client, instproxy_status_cb_t status_cb, const char *operation, void *user_data)
 {
 	instproxy_error_t res = INSTPROXY_E_UNKNOWN_ERROR;
 	if (status_cb) {
@@ -392,15 +412,15 @@ static instproxy_error_t instproxy_create_status_updater(instproxy_client_t clie
 			data->client = client;
 			data->cbfunc = status_cb;
 			data->operation = strdup(operation);
+			data->user_data = user_data;
 
-			client->status_updater = g_thread_create(instproxy_status_updater, data, TRUE, NULL);
-			if (client->status_updater) {
+			if (thread_create(&client->status_updater, instproxy_status_updater, data) == 0) {
 				res = INSTPROXY_E_SUCCESS;
 			}
 		}
 	} else {
 		/* sync mode */
-		res = instproxy_perform_operation(client, NULL, operation);
+		res = instproxy_perform_operation(client, NULL, operation, NULL);
 	}
 	return res;
 }
@@ -415,11 +435,12 @@ static instproxy_error_t instproxy_create_status_updater(instproxy_client_t clie
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
  * @param command The command to execute.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured. 
  */
-static instproxy_error_t instproxy_install_or_upgrade(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb, const char *command)
+static instproxy_error_t instproxy_install_or_upgrade(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb, const char *command, void *user_data)
 {
 	if (!client || !client->parent || !pkg_path) {
 		return INSTPROXY_E_INVALID_ARG;
@@ -437,7 +458,7 @@ static instproxy_error_t instproxy_install_or_upgrade(instproxy_client_t client,
 		return res;
 	}
 
-	return instproxy_create_status_updater(client, status_cb, command);
+	return instproxy_create_status_updater(client, status_cb, command, user_data);
 }
 
 /**
@@ -454,6 +475,7 @@ static instproxy_error_t instproxy_install_or_upgrade(instproxy_client_t client,
  *        an .app directory instead of an install package.
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -463,9 +485,9 @@ static instproxy_error_t instproxy_install_or_upgrade(instproxy_client_t client,
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_install(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_install(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
-	return instproxy_install_or_upgrade(client, pkg_path, client_options, status_cb, "Install");
+	return instproxy_install_or_upgrade(client, pkg_path, client_options, status_cb, "Install", user_data);
 }
 
 /**
@@ -484,6 +506,7 @@ instproxy_error_t instproxy_install(instproxy_client_t client, const char *pkg_p
  *        an .app directory instead of an install package.
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -493,9 +516,9 @@ instproxy_error_t instproxy_install(instproxy_client_t client, const char *pkg_p
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_upgrade(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_upgrade(instproxy_client_t client, const char *pkg_path, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
-	return instproxy_install_or_upgrade(client, pkg_path, client_options, status_cb, "Upgrade");
+	return instproxy_install_or_upgrade(client, pkg_path, client_options, status_cb, "Upgrade", user_data);
 }
 
 /**
@@ -507,6 +530,7 @@ instproxy_error_t instproxy_upgrade(instproxy_client_t client, const char *pkg_p
  *        Currently there are no known client options, so pass NULL here.
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -516,7 +540,7 @@ instproxy_error_t instproxy_upgrade(instproxy_client_t client, const char *pkg_p
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_uninstall(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_uninstall(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
 	if (!client || !client->parent || !appid) {
 		return INSTPROXY_E_INVALID_ARG;
@@ -527,22 +551,17 @@ instproxy_error_t instproxy_uninstall(instproxy_client_t client, const char *app
 	}
 
 	instproxy_error_t res = INSTPROXY_E_UNKNOWN_ERROR;
-	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "ApplicationIdentifier", plist_new_string(appid));
-	plist_dict_insert_item(dict, "Command", plist_new_string("Uninstall"));
 
 	instproxy_lock(client);
 	res = instproxy_send_command(client, "Uninstall", client_options, appid, NULL);
 	instproxy_unlock(client);
 
-	plist_free(dict);
-
 	if (res != INSTPROXY_E_SUCCESS) {
 		debug_info("could not send plist, error %d", res);
 		return res;
 	}
 
-	return instproxy_create_status_updater(client, status_cb, "Uninstall");
+	return instproxy_create_status_updater(client, status_cb, "Uninstall", user_data);
 }
 
 /**
@@ -599,6 +618,7 @@ leave_unlock:
  *          "ArchiveType" -> "ApplicationOnly" 
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -608,7 +628,7 @@ leave_unlock:
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_archive(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_archive(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
 	if (!client || !client->parent || !appid)
 		return INSTPROXY_E_INVALID_ARG;
@@ -625,7 +645,7 @@ instproxy_error_t instproxy_archive(instproxy_client_t client, const char *appid
 		debug_info("could not send plist, error %d", res);
 		return res;
 	}
-	return instproxy_create_status_updater(client, status_cb, "Archive");
+	return instproxy_create_status_updater(client, status_cb, "Archive", user_data);
 }
 
 /**
@@ -639,6 +659,7 @@ instproxy_error_t instproxy_archive(instproxy_client_t client, const char *appid
  *        Currently there are no known client options, so pass NULL here.
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -648,7 +669,7 @@ instproxy_error_t instproxy_archive(instproxy_client_t client, const char *appid
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_restore(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_restore(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
 	if (!client || !client->parent || !appid)
 		return INSTPROXY_E_INVALID_ARG;
@@ -665,7 +686,7 @@ instproxy_error_t instproxy_restore(instproxy_client_t client, const char *appid
 		debug_info("could not send plist, error %d", res);
 		return res;
 	}
-	return instproxy_create_status_updater(client, status_cb, "Restore");
+	return instproxy_create_status_updater(client, status_cb, "Restore", user_data);
 }
 
 /**
@@ -679,6 +700,7 @@ instproxy_error_t instproxy_restore(instproxy_client_t client, const char *appid
  *        Currently there are no known client options, so passing NULL is fine.
  * @param status_cb Callback function for progress and status information. If
  *        NULL is passed, this function will run synchronously.
+ * @param user_data Callback data passed to status_cb.
  *
  * @return INSTPROXY_E_SUCCESS on success or an INSTPROXY_E_* error value if
  *     an error occured.
@@ -688,7 +710,7 @@ instproxy_error_t instproxy_restore(instproxy_client_t client, const char *appid
  *     created successfully; any error occuring during the operation has to be
  *     handled inside the specified callback function.
  */
-instproxy_error_t instproxy_remove_archive(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb)
+instproxy_error_t instproxy_remove_archive(instproxy_client_t client, const char *appid, plist_t client_options, instproxy_status_cb_t status_cb, void *user_data)
 {
 	if (!client || !client->parent || !appid)
 		return INSTPROXY_E_INVALID_ARG;
@@ -705,7 +727,7 @@ instproxy_error_t instproxy_remove_archive(instproxy_client_t client, const char
 		debug_info("could not send plist, error %d", res);
 		return res;
 	}
-	return instproxy_create_status_updater(client, status_cb, "RemoveArchive");
+	return instproxy_create_status_updater(client, status_cb, "RemoveArchive", user_data);
 }
 
 /**
@@ -724,9 +746,9 @@ plist_t instproxy_client_options_new()
  * @param client_options The client options to modify.
  * @param ... KEY, VALUE, [KEY, VALUE], NULL
  *
- * @note The keys and values passed are expected to be strings, except for
- *       "ApplicationSINF" and "iTunesMetadata" expecting a plist node of type
- *       PLIST_DATA as value, or "SkipUninstall" needing int as value.
+ * @note The keys and values passed are expected to be strings, except for the
+ *       keys "ApplicationSINF", "iTunesMetadata", "ReturnAttributes" which are
+ *       expecting a plist_t node as value and "SkipUninstall" expects int.
  */
 void instproxy_client_options_add(plist_t client_options, ...)
 {
@@ -739,21 +761,21 @@ void instproxy_client_options_add(plist_t client_options, ...)
 		char *key = strdup(arg);
 		if (!strcmp(key, "SkipUninstall")) {
 			int intval = va_arg(args, int);
-			plist_dict_insert_item(client_options, key, plist_new_bool(intval));
-		} else if (!strcmp(key, "ApplicationSINF") || !strcmp(key, "iTunesMetadata")) {
+			plist_dict_set_item(client_options, key, plist_new_bool(intval));
+		} else if (!strcmp(key, "ApplicationSINF") || !strcmp(key, "iTunesMetadata") || !strcmp(key, "ReturnAttributes")) {
 			plist_t plistval = va_arg(args, plist_t);
 			if (!plistval) {
 				free(key);
 				break;
 			}
-			plist_dict_insert_item(client_options, key, plist_copy(plistval));
+			plist_dict_set_item(client_options, key, plist_copy(plistval));
 		} else {
 			char *strval = va_arg(args, char*);
 			if (!strval) {
 				free(key);
 				break;
 			}
-			plist_dict_insert_item(client_options, key, plist_new_string(strval));
+			plist_dict_set_item(client_options, key, plist_new_string(strval));
 		}
 		free(key);
 		arg = va_arg(args, char*);
@@ -773,3 +795,112 @@ void instproxy_client_options_free(plist_t client_options)
 		plist_free(client_options);
 	}
 }
+
+/**
+ * Query the device for the path of an application.
+ *
+ * @param client The connected installation proxy client.
+ * @param appid ApplicationIdentifier of app to retrieve the path for.
+ * @param path Pointer to store the device path for the application
+ *        which is set to NULL if it could not be determined.
+ *
+ * @return INSTPROXY_E_SUCCESS on success, INSTPROXY_E_OP_FAILED if
+ *         the path could not be determined or an INSTPROXY_E_* error
+ *         value if an error occured.
+ *
+ * @note This implementation browses all applications and matches the
+ *       right entry by the application identifier.
+ */
+instproxy_error_t instproxy_client_get_path_for_bundle_identifier(instproxy_client_t client, const char* appid, char** path)
+{
+	if (!client || !client->parent || !appid)
+		return INSTPROXY_E_INVALID_ARG;
+
+	plist_t apps = NULL;
+
+	// create client options for any application types
+	plist_t client_opts = instproxy_client_options_new();
+	instproxy_client_options_add(client_opts, "ApplicationType", "Any", NULL);
+
+	// only return attributes we need
+	plist_t return_attributes = plist_new_array();
+	plist_array_append_item(return_attributes, plist_new_string("CFBundleIdentifier"));
+	plist_array_append_item(return_attributes, plist_new_string("CFBundleExecutable"));
+	plist_array_append_item(return_attributes, plist_new_string("Path"));
+	instproxy_client_options_add(client_opts, "ReturnAttributes", return_attributes, NULL);
+	plist_free(return_attributes);
+	return_attributes = NULL;
+
+	// query device for list of apps
+	instproxy_error_t ierr = instproxy_browse(client, client_opts, &apps);
+	instproxy_client_options_free(client_opts);
+	if (ierr != INSTPROXY_E_SUCCESS) {
+		return ierr;
+	}
+
+	plist_t app_found = NULL;
+	uint32_t i;
+	for (i = 0; i < plist_array_get_size(apps); i++) {
+		char *appid_str = NULL;
+		plist_t app_info = plist_array_get_item(apps, i);
+		plist_t idp = plist_dict_get_item(app_info, "CFBundleIdentifier");
+		if (idp) {
+			plist_get_string_val(idp, &appid_str);
+		}
+		if (appid_str && strcmp(appid, appid_str) == 0) {
+			app_found = app_info;
+		}
+		free(appid_str);
+		if (app_found) {
+			break;
+		}
+	}
+
+	if (!app_found) {
+		if (apps)
+			plist_free(apps);
+		*path = NULL;
+		return INSTPROXY_E_OP_FAILED;
+	}
+
+	char* path_str = NULL;
+	plist_t path_p = plist_dict_get_item(app_found, "Path");
+	if (path_p) {
+		plist_get_string_val(path_p, &path_str);
+	}
+
+	char* exec_str = NULL;
+	plist_t exec_p = plist_dict_get_item(app_found, "CFBundleExecutable");
+	if (exec_p) {
+		plist_get_string_val(exec_p, &exec_str);
+	}
+
+	if (!path_str) {
+		debug_info("app path not found");
+		return INSTPROXY_E_OP_FAILED;
+	}
+
+	if (!exec_str) {
+		debug_info("bundle executable not found");
+		return INSTPROXY_E_OP_FAILED;
+	}
+
+	plist_free(apps);
+
+	char* ret = (char*)malloc(strlen(path_str) + 1 + strlen(exec_str) + 1);
+	strcpy(ret, path_str);
+	strcat(ret, "/");
+	strcat(ret, exec_str);
+
+	*path = ret;
+
+	if (path_str) {
+		free(path_str);
+	}
+
+	if (exec_str) {
+		free(exec_str);
+	}
+
+	return INSTPROXY_E_SUCCESS;
+}
