@@ -2,6 +2,8 @@
  * lockdown.c
  * com.apple.mobile.lockdownd service implementation.
  *
+ * Copyright (c) 2014 Martin Szulecki All Rights Reserved.
+ * Copyright (c) 2014 Nikias Bassen. All Rights Reserved.
  * Copyright (c) 2010 Bryan Forbes All Rights Reserved.
  * Copyright (c) 2008 Zach C. All Rights Reserved.
  *
@@ -20,35 +22,44 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <string.h>
 #include <stdlib.h>
 #define _GNU_SOURCE 1
 #define __USE_GNU 1
 #include <stdio.h>
 #include <ctype.h>
-#include <glib.h>
+#include <unistd.h>
+#ifdef HAVE_OPENSSL
+#include <openssl/pem.h>
+#include <openssl/x509.h>
+#include <openssl/x509v3.h>
+#else
 #include <libtasn1.h>
 #include <gnutls/x509.h>
+#include <gnutls/crypto.h>
+#endif
 #include <plist/plist.h>
 
 #include "property_list_service.h"
 #include "lockdown.h"
 #include "idevice.h"
-#include "debug.h"
-#include "userpref.h"
+#include "common/debug.h"
+#include "common/userpref.h"
+#include "common/utils.h"
+#include "asprintf.h"
+
+#ifdef WIN32
+#include <windows.h>
+#define sleep(x) Sleep(x*1000)
+#endif
 
 #define RESULT_SUCCESS 0
 #define RESULT_FAILURE 1
 
-const ASN1_ARRAY_TYPE pkcs1_asn1_tab[] = {
-	{"PKCS1", 536872976, 0},
-	{0, 1073741836, 0},
-	{"RSAPublicKey", 536870917, 0},
-	{"modulus", 1073741827, 0},
-	{"publicExponent", 3, 0},
-	{0, 0, 0}
-};
-
 /**
  * Internally used function for checking the result from lockdown's answer
  * plist to a previously sent request.
@@ -86,7 +97,23 @@ static int lockdown_check_result(plist_t dict, const char *query_match)
 
 	plist_t result_node = plist_dict_get_item(dict, "Result");
 	if (!result_node) {
-		return ret;
+		/* iOS 5: the 'Result' key is not present anymore.
+		   But we need to check for the 'Error' key. */
+		plist_t err_node = plist_dict_get_item(dict, "Error");
+		if (err_node) {
+			if (plist_get_node_type(err_node) == PLIST_STRING) {
+				char *err_value = NULL;
+				plist_get_string_val(err_node, &err_value);
+				if (err_value) {
+					debug_info("ERROR: %s", err_value);
+					free(err_value);
+				} else {
+					debug_info("ERROR: unknown error occured");
+				}
+			}
+			return RESULT_FAILURE;
+		}
+		return RESULT_SUCCESS;
 	}
 
 	plist_type result_type = plist_get_node_type(result_node);
@@ -123,7 +150,7 @@ static void plist_dict_add_label(plist_t plist, const char *label)
 {
 	if (plist && label) {
 		if (plist_get_node_type(plist) == PLIST_DICT)
-			plist_dict_insert_item(plist, "Label", plist_new_string(label));
+			plist_dict_set_item(plist, "Label", plist_new_string(label));
 	}
 }
 
@@ -151,8 +178,8 @@ lockdownd_error_t lockdownd_stop_session(lockdownd_client_t client, const char *
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("StopSession"));
-	plist_dict_insert_item(dict,"SessionID", plist_new_string(session_id));
+	plist_dict_set_item(dict,"Request", plist_new_string("StopSession"));
+	plist_dict_set_item(dict,"SessionID", plist_new_string(session_id));
 
 	debug_info("stopping session %s", session_id);
 
@@ -173,11 +200,50 @@ lockdownd_error_t lockdownd_stop_session(lockdownd_client_t client, const char *
 		debug_info("success");
 		ret = LOCKDOWN_E_SUCCESS;
 	}
+
 	plist_free(dict);
 	dict = NULL;
+
+	if (client->session_id) {
+		free(client->session_id);
+		client->session_id = NULL;
+	}
+
 	if (client->ssl_enabled) {
 		property_list_service_disable_ssl(client->parent);
+		client->ssl_enabled = 0;
 	}
+
+	return ret;
+}
+
+static lockdownd_error_t lockdownd_client_free_simple(lockdownd_client_t client)
+{
+	if (!client)
+		return LOCKDOWN_E_INVALID_ARG;
+
+	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
+
+	if (client->parent) {
+		if (property_list_service_client_free(client->parent) == PROPERTY_LIST_SERVICE_E_SUCCESS) {
+			ret = LOCKDOWN_E_SUCCESS;
+		}
+	}
+
+	if (client->session_id) {
+		free(client->session_id);
+		client->session_id = NULL;
+	}
+	if (client->udid) {
+		free(client->udid);
+	}
+	if (client->label) {
+		free(client->label);
+	}
+
+	free(client);
+	client = NULL;
+
 	return ret;
 }
 
@@ -193,29 +259,15 @@ lockdownd_error_t lockdownd_client_free(lockdownd_client_t client)
 {
 	if (!client)
 		return LOCKDOWN_E_INVALID_ARG;
+
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 
 	if (client->session_id) {
 		lockdownd_stop_session(client, client->session_id);
-		free(client->session_id);
 	}
 
-	if (client->parent) {
-		lockdownd_goodbye(client);
+	ret = lockdownd_client_free_simple(client);
 
-		if (property_list_service_client_free(client->parent) == PROPERTY_LIST_SERVICE_E_SUCCESS) {
-			ret = LOCKDOWN_E_SUCCESS;
-		}
-	}
-
-	if (client->uuid) {
-		free(client->uuid);
-	}
-	if (client->label) {
-		free(client->label);
-	}
-
-	free(client);
 	return ret;
 }
 
@@ -308,7 +360,7 @@ lockdownd_error_t lockdownd_query_type(lockdownd_client_t client, char **type)
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("QueryType"));
+	plist_dict_set_item(dict,"Request", plist_new_string("QueryType"));
 
 	debug_info("called");
 	ret = lockdownd_send(client, dict);
@@ -322,14 +374,21 @@ lockdownd_error_t lockdownd_query_type(lockdownd_client_t client, char **type)
 		return ret;
 
 	ret = LOCKDOWN_E_UNKNOWN_ERROR;
-	if (lockdown_check_result(dict, "QueryType") == RESULT_SUCCESS) {
+	plist_t type_node = plist_dict_get_item(dict, "Type");
+	if (type_node && (plist_get_node_type(type_node) == PLIST_STRING)) {
+		char* typestr = NULL;
+		plist_get_string_val(type_node, &typestr);
+		debug_info("success with type %s", typestr);
 		/* return the type if requested */
 		if (type != NULL) {
-			plist_t type_node = plist_dict_get_item(dict, "Type");
-			plist_get_string_val(type_node, type);
+			*type = typestr;
+		} else {
+			free(typestr);
 		}
-		debug_info("success with type %s", *type);
 		ret = LOCKDOWN_E_SUCCESS;
+	} else {
+		debug_info("hmm. QueryType response does not contain a type?!");
+		debug_plist(dict);
 	}
 	plist_free(dict);
 	dict = NULL;
@@ -359,12 +418,12 @@ lockdownd_error_t lockdownd_get_value(lockdownd_client_t client, const char *dom
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
 	if (domain) {
-		plist_dict_insert_item(dict,"Domain", plist_new_string(domain));
+		plist_dict_set_item(dict,"Domain", plist_new_string(domain));
 	}
 	if (key) {
-		plist_dict_insert_item(dict,"Key", plist_new_string(key));
+		plist_dict_set_item(dict,"Key", plist_new_string(key));
 	}
-	plist_dict_insert_item(dict,"Request", plist_new_string("GetValue"));
+	plist_dict_set_item(dict,"Request", plist_new_string("GetValue"));
 
 	/* send to device */
 	ret = lockdownd_send(client, dict);
@@ -423,13 +482,13 @@ lockdownd_error_t lockdownd_set_value(lockdownd_client_t client, const char *dom
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
 	if (domain) {
-		plist_dict_insert_item(dict,"Domain", plist_new_string(domain));
+		plist_dict_set_item(dict,"Domain", plist_new_string(domain));
 	}
 	if (key) {
-		plist_dict_insert_item(dict,"Key", plist_new_string(key));
+		plist_dict_set_item(dict,"Key", plist_new_string(key));
 	}
-	plist_dict_insert_item(dict,"Request", plist_new_string("SetValue"));
-	plist_dict_insert_item(dict,"Value", value);
+	plist_dict_set_item(dict,"Request", plist_new_string("SetValue"));
+	plist_dict_set_item(dict,"Value", value);
 
 	/* send to device */
 	ret = lockdownd_send(client, dict);
@@ -482,12 +541,12 @@ lockdownd_error_t lockdownd_remove_value(lockdownd_client_t client, const char *
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
 	if (domain) {
-		plist_dict_insert_item(dict,"Domain", plist_new_string(domain));
+		plist_dict_set_item(dict,"Domain", plist_new_string(domain));
 	}
 	if (key) {
-		plist_dict_insert_item(dict,"Key", plist_new_string(key));
+		plist_dict_set_item(dict,"Key", plist_new_string(key));
 	}
-	plist_dict_insert_item(dict,"Request", plist_new_string("RemoveValue"));
+	plist_dict_set_item(dict,"Request", plist_new_string("RemoveValue"));
 
 	/* send to device */
 	ret = lockdownd_send(client, dict);
@@ -521,12 +580,12 @@ lockdownd_error_t lockdownd_remove_value(lockdownd_client_t client, const char *
  * Returns the unique id of the device from lockdownd.
  *
  * @param client An initialized lockdownd client.
- * @param uuid Holds the unique id of the device. The caller is responsible
+ * @param udid Holds the unique id of the device. The caller is responsible
  *  for freeing the memory.
  *
  * @return LOCKDOWN_E_SUCCESS on success
  */
-lockdownd_error_t lockdownd_get_device_uuid(lockdownd_client_t client, char **uuid)
+lockdownd_error_t lockdownd_get_device_udid(lockdownd_client_t client, char **udid)
 {
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 	plist_t value = NULL;
@@ -535,7 +594,7 @@ lockdownd_error_t lockdownd_get_device_uuid(lockdownd_client_t client, char **uu
 	if (ret != LOCKDOWN_E_SUCCESS) {
 		return ret;
 	}
-	plist_get_string_val(value, uuid);
+	plist_get_string_val(value, udid);
 
 	plist_free(value);
 	value = NULL;
@@ -551,7 +610,7 @@ lockdownd_error_t lockdownd_get_device_uuid(lockdownd_client_t client, char **uu
  *
  * @return LOCKDOWN_E_SUCCESS on success
  */
-lockdownd_error_t lockdownd_get_device_public_key(lockdownd_client_t client, gnutls_datum_t * public_key)
+static lockdownd_error_t lockdownd_get_device_public_key_as_key_data(lockdownd_client_t client, key_data_t *public_key)
 {
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 	plist_t value = NULL;
@@ -618,9 +677,14 @@ lockdownd_error_t lockdownd_client_new(idevice_t device, lockdownd_client_t *cli
 	if (!client)
 		return LOCKDOWN_E_INVALID_ARG;
 
+	static struct lockdownd_service_descriptor service = {
+		.port = 0xf27e,
+		.ssl_enabled = 0
+	};
+
 	property_list_service_client_t plistclient = NULL;
-	if (property_list_service_client_new(device, 0xf27e, &plistclient) != PROPERTY_LIST_SERVICE_E_SUCCESS) {
-		debug_info("could not connect to lockdownd (device %s)", device->uuid);
+	if (property_list_service_client_new(device, (lockdownd_service_descriptor_t)&service, &plistclient) != PROPERTY_LIST_SERVICE_E_SUCCESS) {
+		debug_info("could not connect to lockdownd (device %s)", device->udid);
 		return LOCKDOWN_E_MUX_ERROR;
 	}
 
@@ -628,7 +692,12 @@ lockdownd_error_t lockdownd_client_new(idevice_t device, lockdownd_client_t *cli
 	client_loc->parent = plistclient;
 	client_loc->ssl_enabled = 0;
 	client_loc->session_id = NULL;
-	client_loc->uuid = NULL;
+
+	if (idevice_get_udid(device, &client_loc->udid) != IDEVICE_E_SUCCESS) {
+		debug_info("failed to get device udid.");
+	}
+	debug_info("device udid: %s", client_loc->udid);
+
 	client_loc->label = label ? strdup(label) : NULL;
 
 	*client = client_loc;
@@ -677,23 +746,25 @@ lockdownd_error_t lockdownd_client_new_with_handshake(idevice_t device, lockdown
 		if (strcmp("com.apple.mobile.lockdown", type)) {
 			debug_info("Warning QueryType request returned \"%s\".", type);
 		}
-		if (type)
-			free(type);
 	}
+	if (type)
+		free(type);
 
-	ret = idevice_get_uuid(device, &client_loc->uuid);
-	if (LOCKDOWN_E_SUCCESS != ret) {
-		debug_info("failed to get device uuid.");
-	}
-	debug_info("device uuid: %s", client_loc->uuid);
+	plist_t pair_record = NULL;
+	userpref_read_pair_record(client_loc->udid, &pair_record);
 
-	userpref_get_host_id(&host_id);
+	pair_record_get_host_id(pair_record, &host_id);
 	if (LOCKDOWN_E_SUCCESS == ret && !host_id) {
 		ret = LOCKDOWN_E_INVALID_CONF;
 	}
 
-	if (LOCKDOWN_E_SUCCESS == ret && !userpref_has_device_public_key(client_loc->uuid))
+	if (LOCKDOWN_E_SUCCESS == ret && !pair_record) {
+		/* attempt pairing */
 		ret = lockdownd_pair(client_loc, NULL);
+	}
+
+	plist_free(pair_record);
+	pair_record = NULL;
 
 	/* in any case, we need to validate pairing to receive trusted host status */
 	ret = lockdownd_validate_pair(client_loc, NULL);
@@ -701,12 +772,21 @@ lockdownd_error_t lockdownd_client_new_with_handshake(idevice_t device, lockdown
 	/* if not paired yet, let's do it now */
 	if (LOCKDOWN_E_INVALID_HOST_ID == ret) {
 		ret = lockdownd_pair(client_loc, NULL);
+
 		if (LOCKDOWN_E_SUCCESS == ret) {
 			ret = lockdownd_validate_pair(client_loc, NULL);
+		} else if (LOCKDOWN_E_PAIRING_DIALOG_PENDING == ret) {
+			debug_info("Device shows the pairing dialog.");
 		}
 	}
 
 	if (LOCKDOWN_E_SUCCESS == ret) {
+		if (!host_id) {
+			userpref_read_pair_record(client_loc->udid, &pair_record);
+			pair_record_get_host_id(pair_record, &host_id);
+			plist_free(pair_record);
+		}
+
 		ret = lockdownd_start_session(client_loc, host_id, NULL, NULL);
 		if (LOCKDOWN_E_SUCCESS != ret) {
 			debug_info("Session opening failed.");
@@ -740,68 +820,79 @@ static plist_t lockdownd_pair_record_to_plist(lockdownd_pair_record_t pair_recor
 	if (!pair_record)
 		return NULL;
 
-	char *host_id_loc = pair_record->host_id;
-
 	/* setup request plist */
 	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "DeviceCertificate", plist_new_data(pair_record->device_certificate, strlen(pair_record->device_certificate)));
-	plist_dict_insert_item(dict, "HostCertificate", plist_new_data(pair_record->host_certificate, strlen(pair_record->host_certificate)));
-	if (!pair_record->host_id)
-		userpref_get_host_id(&host_id_loc);
-	plist_dict_insert_item(dict, "HostID", plist_new_string(host_id_loc));
-	plist_dict_insert_item(dict, "RootCertificate", plist_new_data(pair_record->root_certificate, strlen(pair_record->root_certificate)));
-
-	if (!pair_record->host_id)
-		free(host_id_loc);
+	plist_dict_set_item(dict, "DeviceCertificate", plist_new_data(pair_record->device_certificate, strlen(pair_record->device_certificate)));
+	plist_dict_set_item(dict, "HostCertificate", plist_new_data(pair_record->host_certificate, strlen(pair_record->host_certificate)));
+	plist_dict_set_item(dict, "HostID", plist_new_string(pair_record->host_id));
+	plist_dict_set_item(dict, "RootCertificate", plist_new_data(pair_record->root_certificate, strlen(pair_record->root_certificate)));
+	plist_dict_set_item(dict, "SystemBUID", plist_new_string(pair_record->system_buid));
 
 	return dict;
 }
 
 /**
- * Generates a new pairing record plist and required certificates for the 
- * supplied public key of the device and the host_id of the caller's host
- * computer.
+ * Generates a pair record plist with required certificates for a specific
+ * device. If a pairing exists, it is loaded from the computer instead of being
+ * generated.
  *
- * @param public_key The public key of the device.
- * @param host_id The HostID to use for the pair record plist.
- * @param pair_record_plist Holds the generated pair record.
+ * @param pair_record_plist Holds the pair record.
  *
  * @return LOCKDOWN_E_SUCCESS on success
  */
-static lockdownd_error_t generate_pair_record_plist(gnutls_datum_t public_key, char *host_id, plist_t *pair_record_plist)
+static lockdownd_error_t pair_record_generate(lockdownd_client_t client, plist_t *pair_record)
 {
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 
-	gnutls_datum_t device_cert = { NULL, 0 };
-	gnutls_datum_t host_cert = { NULL, 0 };
-	gnutls_datum_t root_cert = { NULL, 0 };
+	key_data_t public_key = { NULL, 0 };
+	char* host_id = NULL;
+	char* system_buid = NULL;
 
-	ret = lockdownd_gen_pair_cert(public_key, &device_cert, &host_cert, &root_cert);
+	/* retrieve device public key */
+	ret = lockdownd_get_device_public_key_as_key_data(client, &public_key);
 	if (ret != LOCKDOWN_E_SUCCESS) {
-		return ret;
+		debug_info("device refused to send public key.");
+		goto leave;
+	}
+	debug_info("device public key follows:\n%.*s", public_key.size, public_key.data);
+
+	*pair_record = plist_new_dict();
+
+	/* generate keys and certificates into pair record */
+	userpref_error_t uret = USERPREF_E_SUCCESS;
+	uret = pair_record_generate_keys_and_certs(*pair_record, public_key);
+	switch(uret) {
+		case USERPREF_E_INVALID_ARG:
+			ret = LOCKDOWN_E_INVALID_ARG;
+			break;
+		case USERPREF_E_INVALID_CONF:
+			ret = LOCKDOWN_E_INVALID_CONF;
+			break;
+		case USERPREF_E_SSL_ERROR:
+			ret = LOCKDOWN_E_SSL_ERROR;
+		default:
+			break;
 	}
 
-	char *host_id_loc = host_id;
+	/* set SystemBUID */
+	userpref_read_system_buid(&system_buid);
+	plist_dict_set_item(*pair_record, USERPREF_SYSTEM_BUID_KEY, plist_new_string(system_buid));
 
-	if (!host_id)
-		userpref_get_host_id(&host_id_loc);
+	/* set HostID */
+	host_id = generate_uuid();
+	pair_record_set_host_id(*pair_record, host_id);
 
-	/* setup request plist */
-	*pair_record_plist = plist_new_dict();
-	plist_dict_insert_item(*pair_record_plist, "DeviceCertificate", plist_new_data((const char*)device_cert.data, device_cert.size));
-	plist_dict_insert_item(*pair_record_plist, "HostCertificate", plist_new_data((const char*)host_cert.data, host_cert.size));
-	plist_dict_insert_item(*pair_record_plist, "HostID", plist_new_string(host_id_loc));
-	plist_dict_insert_item(*pair_record_plist, "RootCertificate", plist_new_data((const char*)root_cert.data, root_cert.size));
+	if (ret != LOCKDOWN_E_SUCCESS) {
+		goto leave;
+	}
 
-	if (!host_id)
-		free(host_id_loc);
-
-	if (device_cert.data)
-		free(device_cert.data);
-	if (host_cert.data)
-		free(host_cert.data);
-	if (root_cert.data)
-		free(root_cert.data);
+leave:
+	if (host_id)
+		free(host_id);
+	if (system_buid)
+		free(system_buid);
+	if (public_key.data)
+		free(public_key.data);
 
 	return ret;
 }
@@ -828,57 +919,81 @@ static lockdownd_error_t lockdownd_do_pair(lockdownd_client_t client, lockdownd_
 
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 	plist_t dict = NULL;
-	plist_t dict_record = NULL;
-	gnutls_datum_t public_key = { NULL, 0 };
+	plist_t pair_record_plist = NULL;
+	plist_t wifi_node = NULL;
 	int pairing_mode = 0; /* 0 = libimobiledevice, 1 = external */
 
-	if (pair_record && pair_record->host_id) {
+	if (pair_record && pair_record->system_buid && pair_record->host_id) {
 		/* valid pair_record passed? */
 		if (!pair_record->device_certificate || !pair_record->host_certificate || !pair_record->root_certificate) {
 			return LOCKDOWN_E_PLIST_ERROR;
 		}
 
 		/* use passed pair_record */
-		dict_record = lockdownd_pair_record_to_plist(pair_record);
+		pair_record_plist = lockdownd_pair_record_to_plist(pair_record);
 
 		pairing_mode = 1;
 	} else {
-		ret = lockdownd_get_device_public_key(client, &public_key);
-		if (ret != LOCKDOWN_E_SUCCESS) {
-			if (public_key.data)
-				free(public_key.data);
-			debug_info("device refused to send public key.");
-			return ret;
-		}
-		debug_info("device public key follows:\n%s", public_key.data);
-		/* get libimobiledevice pair_record */
-		ret = generate_pair_record_plist(public_key, NULL, &dict_record);
-		if (ret != LOCKDOWN_E_SUCCESS) {
-			if (dict_record)
-				plist_free(dict_record);
-			return ret;
+		/* generate a new pair record if pairing */
+		if (!strcmp("Pair", verb)) {
+			ret = pair_record_generate(client, &pair_record_plist);
+
+			if (ret != LOCKDOWN_E_SUCCESS) {
+				if (pair_record_plist)
+					plist_free(pair_record_plist);
+				return ret;
+			}
+
+			/* get wifi mac now, if we get it later we fail on iOS 7 which causes a reconnect */
+			lockdownd_get_value(client, NULL, "WiFiAddress", &wifi_node);
+		} else {
+			/* use existing pair record */
+			if (userpref_has_pair_record(client->udid)) {
+				userpref_read_pair_record(client->udid, &pair_record_plist);
+			} else {
+				return LOCKDOWN_E_INVALID_HOST_ID;
+			}
 		}
 	}
 
-	/* Setup Pair request plist */
+	plist_t request_pair_record = plist_copy(pair_record_plist);
+
+	/* remove stuff that is private */
+	plist_dict_remove_item(request_pair_record, USERPREF_ROOT_PRIVATE_KEY_KEY);
+	plist_dict_remove_item(request_pair_record, USERPREF_HOST_PRIVATE_KEY_KEY);
+
+	/* setup pair request plist */
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"PairRecord", dict_record);
-	plist_dict_insert_item(dict, "Request", plist_new_string(verb));
+	plist_dict_set_item(dict, "PairRecord", request_pair_record);
+	plist_dict_set_item(dict, "Request", plist_new_string(verb));
+	plist_dict_set_item(dict, "ProtocolVersion", plist_new_string(LOCKDOWN_PROTOCOL_VERSION));
+
+	plist_t options = plist_new_dict();
+	plist_dict_set_item(options, "ExtendedPairingErrors", plist_new_bool(1));
+	plist_dict_set_item(dict, "PairingOptions", options);
 
 	/* send to device */
 	ret = lockdownd_send(client, dict);
 	plist_free(dict);
 	dict = NULL;
 
-	if (ret != LOCKDOWN_E_SUCCESS)
+	if (ret != LOCKDOWN_E_SUCCESS) {
+		plist_free(pair_record_plist);
+		if (wifi_node)
+			plist_free(wifi_node);
 		return ret;
+	}
 
 	/* Now get device's answer */
 	ret = lockdownd_receive(client, &dict);
 
-	if (ret != LOCKDOWN_E_SUCCESS)
+	if (ret != LOCKDOWN_E_SUCCESS) {
+		plist_free(pair_record_plist);
+		if (wifi_node)
+			plist_free(wifi_node);
 		return ret;
+	}
 
 	if (strcmp(verb, "Unpair") == 0) {
 		/* workaround for Unpair giving back ValidatePair,
@@ -896,13 +1011,34 @@ static lockdownd_error_t lockdownd_do_pair(lockdownd_client_t client, lockdownd_
 	if (ret == LOCKDOWN_E_SUCCESS) {
 		debug_info("%s success", verb);
 		if (!pairing_mode) {
+			debug_info("internal pairing mode");
 			if (!strcmp("Unpair", verb)) {
 				/* remove public key from config */
-				userpref_remove_device_public_key(client->uuid);
+				userpref_delete_pair_record(client->udid);
 			} else {
-				/* store public key in config */
-				userpref_set_device_public_key(client->uuid, public_key);
+				if (!strcmp("Pair", verb)) {
+					/* add returned escrow bag if available */
+					plist_t extra_node = plist_dict_get_item(dict, USERPREF_ESCROW_BAG_KEY);
+					if (extra_node && plist_get_node_type(extra_node) == PLIST_DATA) {
+						debug_info("Saving EscrowBag from response in pair record");
+						plist_dict_set_item(pair_record_plist, USERPREF_ESCROW_BAG_KEY, plist_copy(extra_node));
+						plist_free(extra_node);
+						extra_node = NULL;
+					}
+
+					/* save previously retrieved wifi mac address in pair record */
+					if (wifi_node) {
+						debug_info("Saving WiFiAddress from device in pair record");
+						plist_dict_set_item(pair_record_plist, USERPREF_WIFI_MAC_ADDRESS_KEY, plist_copy(wifi_node));
+						plist_free(wifi_node);
+						wifi_node = NULL;
+					}
+
+					userpref_save_pair_record(client->udid, pair_record_plist);
+				}
 			}
+		} else {
+			debug_info("external pairing mode");
 		}
 	} else {
 		debug_info("%s failure", verb);
@@ -918,6 +1054,10 @@ static lockdownd_error_t lockdownd_do_pair(lockdownd_client_t client, lockdownd_
 					ret = LOCKDOWN_E_PASSWORD_PROTECTED;
 				} else if (!strcmp(value, "InvalidHostID")) {
 					ret = LOCKDOWN_E_INVALID_HOST_ID;
+				} else if (!strcmp(value, "UserDeniedPairing")) {
+					ret = LOCKDOWN_E_USER_DENIED_PAIRING;
+				} else if (!strcmp(value, "PairingDialogResponsePending")) {
+					ret = LOCKDOWN_E_PAIRING_DIALOG_PENDING;
 				}
 				free(value);
 			}
@@ -926,10 +1066,20 @@ static lockdownd_error_t lockdownd_do_pair(lockdownd_client_t client, lockdownd_
 			error_node = NULL;
 		}
 	}
+
+	if (pair_record_plist) {
+		plist_free(pair_record_plist);
+		pair_record_plist = NULL;
+	}
+
+	if (wifi_node) {
+		plist_free(wifi_node);
+		wifi_node = NULL;
+	}
+
 	plist_free(dict);
 	dict = NULL;
-	if (public_key.data)
-		free(public_key.data);
+
 	return ret;
 }
 
@@ -1009,7 +1159,7 @@ lockdownd_error_t lockdownd_enter_recovery(lockdownd_client_t client)
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("EnterRecovery"));
+	plist_dict_set_item(dict,"Request", plist_new_string("EnterRecovery"));
 
 	debug_info("telling device to enter recovery mode");
 
@@ -1033,8 +1183,9 @@ lockdownd_error_t lockdownd_enter_recovery(lockdownd_client_t client)
  *
  * @param client The lockdown client
  *
- * @return LOCKDOWN_E_SUCCESS on success, NP_E_INVALID_ARG when client is NULL,
- *  LOCKDOWN_E_PLIST_ERROR if the device did not acknowledge the request
+ * @return LOCKDOWN_E_SUCCESS on success, LOCKDOWN_E_INVALID_ARG when client
+ *  is NULL, LOCKDOWN_E_PLIST_ERROR if the device did not acknowledge the
+ *  request
  */
 lockdownd_error_t lockdownd_goodbye(lockdownd_client_t client)
 {
@@ -1045,7 +1196,7 @@ lockdownd_error_t lockdownd_goodbye(lockdownd_client_t client)
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("Goodbye"));
+	plist_dict_set_item(dict,"Request", plist_new_string("Goodbye"));
 
 	debug_info("called");
 
@@ -1065,164 +1216,6 @@ lockdownd_error_t lockdownd_goodbye(lockdownd_client_t client)
 	}
 	plist_free(dict);
 	dict = NULL;
-	return ret;
-}
-
-/**
- * Generates the device certificate from the public key as well as the host
- * and root certificates.
- *
- * @param public_key The public key of the device to use for generation.
- * @param odevice_cert Holds the generated device certificate.
- * @param ohost_cert Holds the generated host certificate.
- * @param oroot_cert Holds the generated root certificate.
- *
- * @return LOCKDOWN_E_SUCCESS on success, NP_E_INVALID_ARG when a parameter is NULL,
- *  LOCKDOWN_E_INVALID_CONF if the internal configuration system failed,
- *  LOCKDOWN_E_SSL_ERROR if the certificates could not be generated
- */
-lockdownd_error_t lockdownd_gen_pair_cert(gnutls_datum_t public_key, gnutls_datum_t * odevice_cert,
-									   gnutls_datum_t * ohost_cert, gnutls_datum_t * oroot_cert)
-{
-	if (!public_key.data || !odevice_cert || !ohost_cert || !oroot_cert)
-		return LOCKDOWN_E_INVALID_ARG;
-	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
-	userpref_error_t uret = USERPREF_E_UNKNOWN_ERROR;
-
-	gnutls_datum_t modulus = { NULL, 0 };
-	gnutls_datum_t exponent = { NULL, 0 };
-
-	/* now decode the PEM encoded key */
-	gnutls_datum_t der_pub_key;
-	if (GNUTLS_E_SUCCESS == gnutls_pem_base64_decode_alloc("RSA PUBLIC KEY", &public_key, &der_pub_key)) {
-
-		/* initalize asn.1 parser */
-		ASN1_TYPE pkcs1 = ASN1_TYPE_EMPTY;
-		if (ASN1_SUCCESS == asn1_array2tree(pkcs1_asn1_tab, &pkcs1, NULL)) {
-
-			ASN1_TYPE asn1_pub_key = ASN1_TYPE_EMPTY;
-			asn1_create_element(pkcs1, "PKCS1.RSAPublicKey", &asn1_pub_key);
-
-			if (ASN1_SUCCESS == asn1_der_decoding(&asn1_pub_key, der_pub_key.data, der_pub_key.size, NULL)) {
-
-				/* get size to read */
-				int ret1 = asn1_read_value(asn1_pub_key, "modulus", NULL, (int*)&modulus.size);
-				int ret2 = asn1_read_value(asn1_pub_key, "publicExponent", NULL, (int*)&exponent.size);
-
-				modulus.data = gnutls_malloc(modulus.size);
-				exponent.data = gnutls_malloc(exponent.size);
-
-				ret1 = asn1_read_value(asn1_pub_key, "modulus", modulus.data, (int*)&modulus.size);
-				ret2 = asn1_read_value(asn1_pub_key, "publicExponent", exponent.data, (int*)&exponent.size);
-				if (ASN1_SUCCESS == ret1 && ASN1_SUCCESS == ret2)
-					ret = LOCKDOWN_E_SUCCESS;
-			}
-			if (asn1_pub_key)
-				asn1_delete_structure(&asn1_pub_key);
-		}
-		if (pkcs1)
-			asn1_delete_structure(&pkcs1);
-	}
-
-	/* now generate certificates */
-	if (LOCKDOWN_E_SUCCESS == ret && 0 != modulus.size && 0 != exponent.size) {
-
-		gnutls_global_init();
-		gnutls_datum_t essentially_null = { (unsigned char*)strdup("abababababababab"), strlen("abababababababab") };
-
-		gnutls_x509_privkey_t fake_privkey, root_privkey, host_privkey;
-		gnutls_x509_crt_t dev_cert, root_cert, host_cert;
-
-		gnutls_x509_privkey_init(&fake_privkey);
-		gnutls_x509_crt_init(&dev_cert);
-		gnutls_x509_crt_init(&root_cert);
-		gnutls_x509_crt_init(&host_cert);
-
-		if (GNUTLS_E_SUCCESS ==
-			gnutls_x509_privkey_import_rsa_raw(fake_privkey, &modulus, &exponent, &essentially_null, &essentially_null,
-											   &essentially_null, &essentially_null)) {
-
-			gnutls_x509_privkey_init(&root_privkey);
-			gnutls_x509_privkey_init(&host_privkey);
-
-			uret = userpref_get_keys_and_certs(root_privkey, root_cert, host_privkey, host_cert);
-
-			if (USERPREF_E_SUCCESS == uret) {
-				/* generate device certificate */
-				gnutls_x509_crt_set_key(dev_cert, fake_privkey);
-				gnutls_x509_crt_set_serial(dev_cert, "\x00", 1);
-				gnutls_x509_crt_set_version(dev_cert, 3);
-				gnutls_x509_crt_set_ca_status(dev_cert, 0);
-				gnutls_x509_crt_set_activation_time(dev_cert, time(NULL));
-				gnutls_x509_crt_set_expiration_time(dev_cert, time(NULL) + (60 * 60 * 24 * 365 * 10));
-				gnutls_x509_crt_sign(dev_cert, root_cert, root_privkey);
-
-				if (LOCKDOWN_E_SUCCESS == ret) {
-					/* if everything went well, export in PEM format */
-					size_t export_size = 0;
-					gnutls_datum_t dev_pem = { NULL, 0 };
-					gnutls_x509_crt_export(dev_cert, GNUTLS_X509_FMT_PEM, NULL, &export_size);
-					dev_pem.data = gnutls_malloc(export_size);
-					gnutls_x509_crt_export(dev_cert, GNUTLS_X509_FMT_PEM, dev_pem.data, &export_size);
-					dev_pem.size = export_size;
-
-					gnutls_datum_t pem_root_cert = { NULL, 0 };
-					gnutls_datum_t pem_host_cert = { NULL, 0 };
-
-					uret = userpref_get_certs_as_pem(&pem_root_cert, &pem_host_cert);
-
-					if (USERPREF_E_SUCCESS == uret) {
-						/* copy buffer for output */
-						odevice_cert->data = malloc(dev_pem.size);
-						memcpy(odevice_cert->data, dev_pem.data, dev_pem.size);
-						odevice_cert->size = dev_pem.size;
-
-						ohost_cert->data = malloc(pem_host_cert.size);
-						memcpy(ohost_cert->data, pem_host_cert.data, pem_host_cert.size);
-						ohost_cert->size = pem_host_cert.size;
-
-						oroot_cert->data = malloc(pem_root_cert.size);
-						memcpy(oroot_cert->data, pem_root_cert.data, pem_root_cert.size);
-						oroot_cert->size = pem_root_cert.size;
-
-						g_free(pem_root_cert.data);
-						g_free(pem_host_cert.data);
-
-						if (dev_pem.data)
-							gnutls_free(dev_pem.data);
-					}
-				}
-			}
-
-			switch(uret) {
-			case USERPREF_E_INVALID_ARG:
-				ret = LOCKDOWN_E_INVALID_ARG;
-				break;
-			case USERPREF_E_INVALID_CONF:
-				ret = LOCKDOWN_E_INVALID_CONF;
-				break;
-			case USERPREF_E_SSL_ERROR:
-				ret = LOCKDOWN_E_SSL_ERROR;
-			default:
-				break;
-			}
-		}
-
-		if (essentially_null.data)
-			free(essentially_null.data);
-		gnutls_x509_crt_deinit(dev_cert);
-		gnutls_x509_crt_deinit(root_cert);
-		gnutls_x509_crt_deinit(host_cert);
-		gnutls_x509_privkey_deinit(fake_privkey);
-		gnutls_x509_privkey_deinit(root_privkey);
-		gnutls_x509_privkey_deinit(host_privkey);
-
-	}
-
-	gnutls_free(modulus.data);
-	gnutls_free(exponent.data);
-
-	gnutls_free(der_pub_key.data);
 
 	return ret;
 }
@@ -1235,8 +1228,8 @@ lockdownd_error_t lockdownd_gen_pair_cert(gnutls_datum_t public_key, gnutls_datu
  * @param session_id The new session_id of the created session
  * @param ssl_enabled Whether SSL communication is used in the session
  *
- * @return LOCKDOWN_E_SUCCESS on success, NP_E_INVALID_ARG when a client or
- *  host_id is NULL, LOCKDOWN_E_PLIST_ERROR if the response plist had errors,
+ * @return LOCKDOWN_E_SUCCESS on success, LOCKDOWN_E_INVALID_ARG when a client
+ *  or host_id is NULL, LOCKDOWN_E_PLIST_ERROR if the response plist had errors,
  *  LOCKDOWN_E_INVALID_HOST_ID if the device does not know the supplied HostID,
  *  LOCKDOWN_E_SSL_ERROR if enabling SSL communication failed
  */
@@ -1251,14 +1244,28 @@ lockdownd_error_t lockdownd_start_session(lockdownd_client_t client, const char
 	/* if we have a running session, stop current one first */
 	if (client->session_id) {
 		lockdownd_stop_session(client, client->session_id);
-		free(client->session_id);
 	}
 
 	/* setup request plist */
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"HostID", plist_new_string(host_id));
-	plist_dict_insert_item(dict,"Request", plist_new_string("StartSession"));
+	plist_dict_set_item(dict,"Request", plist_new_string("StartSession"));
+
+	/* add host id */
+	if (host_id) {
+		plist_dict_set_item(dict, "HostID", plist_new_string(host_id));
+	}
+
+	/* add system buid */
+	char *system_buid = NULL;
+	userpref_read_system_buid(&system_buid);
+	if (system_buid) {
+		plist_dict_set_item(dict, "SystemBUID", plist_new_string(system_buid));
+		if (system_buid) {
+			free(system_buid);
+			system_buid = NULL;
+		}
+	}
 
 	ret = lockdownd_send(client, dict);
 	plist_free(dict);
@@ -1299,6 +1306,7 @@ lockdownd_error_t lockdownd_start_session(lockdownd_client_t client, const char
 		if (session_node && (plist_get_node_type(session_node) == PLIST_STRING)) {
 			plist_get_string_val(session_node, &client->session_id);
 		}
+
 		if (client->session_id) {
 			debug_info("SessionID: %s", client->session_id);
 			if (session_id != NULL)
@@ -1306,7 +1314,9 @@ lockdownd_error_t lockdownd_start_session(lockdownd_client_t client, const char
 		} else {
 			debug_info("Failed to get SessionID!");
 		}
+
 		debug_info("Enable SSL Session: %s", (use_ssl?"true":"false"));
+
 		if (use_ssl) {
 			ret = property_list_service_enable_ssl(client->parent);
 			if (ret == PROPERTY_LIST_SERVICE_E_SUCCESS) {
@@ -1331,37 +1341,33 @@ lockdownd_error_t lockdownd_start_session(lockdownd_client_t client, const char
  * Requests to start a service and retrieve it's port on success.
  *
  * @param client The lockdownd client
- * @param service The name of the service to start
- * @param port The port number the service was started on
+ * @param identifier The identifier of the service to start
+ * @param descriptor The service descriptor on success or NULL on failure
  
  * @return LOCKDOWN_E_SUCCESS on success, NP_E_INVALID_ARG if a parameter
  *  is NULL, LOCKDOWN_E_INVALID_SERVICE if the requested service is not known
  *  by the device, LOCKDOWN_E_START_SERVICE_FAILED if the service could not because
  *  started by the device
  */
-lockdownd_error_t lockdownd_start_service(lockdownd_client_t client, const char *service, uint16_t *port)
+lockdownd_error_t lockdownd_start_service(lockdownd_client_t client, const char *identifier, lockdownd_service_descriptor_t *service)
 {
-	if (!client || !service || !port)
+	if (!client || !identifier || !service)
 		return LOCKDOWN_E_INVALID_ARG;
 
-	char *host_id = NULL;
-	userpref_get_host_id(&host_id);
-	if (!host_id)
-		return LOCKDOWN_E_INVALID_CONF;
-	if (!client->session_id)
-		return LOCKDOWN_E_NO_RUNNING_SESSION;
+	if (*service) {
+		// reset fields if service descriptor is reused
+		(*service)->port = 0;
+		(*service)->ssl_enabled = 0;
+	}
 
 	plist_t dict = NULL;
 	uint16_t port_loc = 0;
 	lockdownd_error_t ret = LOCKDOWN_E_UNKNOWN_ERROR;
 
-	free(host_id);
-	host_id = NULL;
-
 	dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("StartService"));
-	plist_dict_insert_item(dict,"Service", plist_new_string(service));
+	plist_dict_set_item(dict,"Request", plist_new_string("StartService"));
+	plist_dict_set_item(dict,"Service", plist_new_string(identifier));
 
 	/* send to device */
 	ret = lockdownd_send(client, dict);
@@ -1381,18 +1387,32 @@ lockdownd_error_t lockdownd_start_service(lockdownd_client_t client, const char
 
 	ret = LOCKDOWN_E_UNKNOWN_ERROR;
 	if (lockdown_check_result(dict, "StartService") == RESULT_SUCCESS) {
-		plist_t port_value_node = plist_dict_get_item(dict, "Port");
+		if (*service == NULL)
+			*service = (lockdownd_service_descriptor_t)malloc(sizeof(struct lockdownd_service_descriptor));
+		(*service)->port = 0;
+		(*service)->ssl_enabled = 0;
 
-		if (port_value_node && (plist_get_node_type(port_value_node) == PLIST_UINT)) {
+		/* read service port number */
+		plist_t node = plist_dict_get_item(dict, "Port");
+		if (node && (plist_get_node_type(node) == PLIST_UINT)) {
 			uint64_t port_value = 0;
-			plist_get_uint_val(port_value_node, &port_value);
+			plist_get_uint_val(node, &port_value);
 
 			if (port_value) {
 				port_loc = port_value;
 				ret = LOCKDOWN_E_SUCCESS;
 			}
-			if (port && ret == LOCKDOWN_E_SUCCESS)
-				*port = port_loc;
+			if (port_loc && ret == LOCKDOWN_E_SUCCESS) {
+				(*service)->port = port_loc;
+			}
+		}
+
+		/* check if the service requires SSL */
+		node = plist_dict_get_item(dict, "EnableServiceSSL");
+		if (node && (plist_get_node_type(node) == PLIST_BOOLEAN)) {
+			uint8_t b = 0;
+			plist_get_bool_val(node, &b);
+			(*service)->ssl_enabled = b;
 		}
 	} else {
 		ret = LOCKDOWN_E_START_SERVICE_FAILED;
@@ -1402,6 +1422,8 @@ lockdownd_error_t lockdownd_start_service(lockdownd_client_t client, const char
 			plist_get_string_val(error_node, &error);
 			if (!strcmp(error, "InvalidService")) {
 				ret = LOCKDOWN_E_INVALID_SERVICE;
+			} else if (!strcmp(error, "NoRunningSession")) {
+				ret = LOCKDOWN_E_NO_RUNNING_SESSION;
 			}
 			free(error);
 		}
@@ -1444,8 +1466,8 @@ lockdownd_error_t lockdownd_activate(lockdownd_client_t client, plist_t activati
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("Activate"));
-	plist_dict_insert_item(dict,"ActivationRecord", plist_copy(activation_record));
+	plist_dict_set_item(dict,"Request", plist_new_string("Activate"));
+	plist_dict_set_item(dict,"ActivationRecord", plist_copy(activation_record));
 
 	ret = lockdownd_send(client, dict);
 	plist_free(dict);
@@ -1502,7 +1524,7 @@ lockdownd_error_t lockdownd_deactivate(lockdownd_client_t client)
 
 	plist_t dict = plist_new_dict();
 	plist_dict_add_label(dict, client->label);
-	plist_dict_insert_item(dict,"Request", plist_new_string("Deactivate"));
+	plist_dict_set_item(dict,"Request", plist_new_string("Deactivate"));
 
 	ret = lockdownd_send(client, dict);
 	plist_free(dict);
@@ -1622,3 +1644,18 @@ lockdownd_error_t lockdownd_data_classes_free(char **classes)
 	}
 	return LOCKDOWN_E_SUCCESS;
 }
+
+/**
+ * Frees memory of a service descriptor as returned by lockdownd_start_service()
+ *
+ * @param sevice A service descriptor instance to free.
+ *
+ * @return LOCKDOWN_E_SUCCESS on success
+ */
+lockdownd_error_t lockdownd_service_descriptor_free(lockdownd_service_descriptor_t service)
+{
+	if (service)
+		free(service);
+
+	return LOCKDOWN_E_SUCCESS;
+}
