@@ -27,24 +27,23 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
-#include <glib.h>
 
 #include "mobilesync.h"
 #include "device_link_service.h"
-#include "debug.h"
+#include "common/debug.h"
 
-#define MSYNC_VERSION_INT1 100
+#define MSYNC_VERSION_INT1 300
 #define MSYNC_VERSION_INT2 100
 
 #define EMPTY_PARAMETER_STRING "___EmptyParameterString___"
 
 /**
- * Convert an device_link_service_error_t value to an mobilesync_error_t value.
+ * Convert an #device_link_service_error_t value to an #mobilesync_error_t value.
  * Used internally to get correct error codes when using device_link_service stuff.
  *
- * @param err An device_link_service_error_t error code
+ * @param err A #device_link_service_error_t error code
  *
- * @return A matching mobilesync_error_t error code,
+ * @return A matching #mobilesync_error_t error code,
  *     MOBILESYNC_E_UNKNOWN_ERROR otherwise.
  */
 static mobilesync_error_t mobilesync_error(device_link_service_error_t err)
@@ -70,22 +69,23 @@ static mobilesync_error_t mobilesync_error(device_link_service_error_t err)
  * Connects to the mobilesync service on the specified device.
  *
  * @param device The device to connect to.
- * @param port Destination port (usually given by lockdownd_start_service).
+ * @param service The service descriptor returned by lockdownd_start_service.
  * @param client Pointer that will be set to a newly allocated
- *     mobilesync_client_t upon successful return.
+ *     #mobilesync_client_t upon successful return.
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID ARG if one
- *     or more parameters are invalid, or DEVICE_LINK_SERVICE_E_BAD_VERSION if
- *     the mobilesync version on the device is newer.
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one or more parameters are invalid
+ * @retval DEVICE_LINK_SERVICE_E_BAD_VERSION if the mobilesync version on
+ * the device is newer.
  */
-mobilesync_error_t mobilesync_client_new(idevice_t device, uint16_t port,
+mobilesync_error_t mobilesync_client_new(idevice_t device, lockdownd_service_descriptor_t service,
 						   mobilesync_client_t * client)
 {
-	if (!device || port == 0 || !client || *client)
+	if (!device || !service || service->port == 0 || !client || *client)
 		return MOBILESYNC_E_INVALID_ARG;
 
 	device_link_service_client_t dlclient = NULL;
-	mobilesync_error_t ret = mobilesync_error(device_link_service_client_new(device, port, &dlclient));
+	mobilesync_error_t ret = mobilesync_error(device_link_service_client_new(device, service, &dlclient));
 	if (ret != MOBILESYNC_E_SUCCESS) {
 		return ret;
 	}
@@ -108,20 +108,40 @@ mobilesync_error_t mobilesync_client_new(idevice_t device, uint16_t port,
 	return ret;
 }
 
+/**
+ * Starts a new mobilesync service on the specified device and connects to it.
+ *
+ * @param device The device to connect to.
+ * @param client Pointer that will point to a newly allocated
+ *     mobilesync_client_t upon successful return. Must be freed using
+ *     mobilesync_client_free() after use.
+ * @param label The label to use for communication. Usually the program name.
+ *  Pass NULL to disable sending the label in requests to lockdownd.
+ *
+ * @return MOBILESYNC_E_SUCCESS on success, or an MOBILESYNC_E_* error
+ *     code otherwise.
+ */
+mobilesync_error_t mobilesync_client_start_service(idevice_t device, mobilesync_client_t * client, const char* label)
+{
+	mobilesync_error_t err = MOBILESYNC_E_UNKNOWN_ERROR;
+	service_client_factory_start_service(device, MOBILESYNC_SERVICE_NAME, (void**)client, label, SERVICE_CONSTRUCTOR(mobilesync_client_new), &err);
+	return err;
+}
+
 /**
  * Disconnects a mobilesync client from the device and frees up the
  * mobilesync client data.
  *
  * @param client The mobilesync client to disconnect and free.
  *
- * @return MOBILESYNC_E_SUCCESS on success, or MOBILESYNC_E_INVALID_ARG
- *     if client is NULL.
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if \a client is NULL.
  */
 mobilesync_error_t mobilesync_client_free(mobilesync_client_t client)
 {
 	if (!client)
 		return MOBILESYNC_E_INVALID_ARG;
-	device_link_service_disconnect(client->parent);
+	device_link_service_disconnect(client->parent, "All done, thanks for the memories");
 	mobilesync_error_t err = mobilesync_error(device_link_service_client_free(client->parent));
 	free(client);
 	return err;
@@ -172,14 +192,16 @@ mobilesync_error_t mobilesync_send(mobilesync_client_t client, plist_t plist)
  * @param computer_data_class_version The version of the data class storage on the computer
  * @param sync_type A pointer to store the sync type reported by the device_anchor
  * @param device_data_class_version The version of the data class storage on the device
+ * @param error_description A pointer to store an error message if reported by the device
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_PLIST_ERROR if
- *    the received plist is not of valid form, MOBILESYNC_E_SYNC_REFUSED if the 
- *    device refused to sync or MOBILESYNC_E_CANCELLED if the device explicitly
- *    cancelled the sync request
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_PLIST_ERROR if the received plist is not of valid form
+ * @retval MOBILESYNC_E_SYNC_REFUSED if the device refused to sync
+ * @retval MOBILESYNC_E_CANCELLED if the device explicitly cancelled the
+ * sync request
  */
-mobilesync_error_t mobilesync_start(mobilesync_client_t client, const char *data_class, mobilesync_anchors_t anchors, uint64_t computer_data_class_version, mobilesync_sync_type_t *sync_type, uint64_t *device_data_class_version)
+mobilesync_error_t mobilesync_start(mobilesync_client_t client, const char *data_class, mobilesync_anchors_t anchors, uint64_t computer_data_class_version, mobilesync_sync_type_t *sync_type, uint64_t *device_data_class_version, char** error_description)
 {
 	if (!client || client->data_class || !data_class ||
 		!anchors || !anchors->computer_anchor) {
@@ -192,6 +214,8 @@ mobilesync_error_t mobilesync_start(mobilesync_client_t client, const char *data
 	plist_t msg = NULL;
 	plist_t response_type_node = NULL;
 
+	*error_description = NULL;
+
 	msg = plist_new_array();
 	plist_array_append_item(msg, plist_new_string("SDMessageSyncDataClassWithDevice"));
 	plist_array_append_item(msg, plist_new_string(data_class));
@@ -231,23 +255,19 @@ mobilesync_error_t mobilesync_start(mobilesync_client_t client, const char *data
 		goto out;
 	}
 
+	// did the device refuse to sync with the computer?
 	if (!strcmp(response_type, "SDMessageRefuseToSyncDataClassWithComputer")) {
-		char *reason = NULL;
 		err = MOBILESYNC_E_SYNC_REFUSED;
-		plist_get_string_val(plist_array_get_item(msg, 2), &reason);
-		debug_info("Device refused sync: %s", reason);
-		free(reason);
-		reason = NULL;
+		plist_get_string_val(plist_array_get_item(msg, 2), error_description);
+		debug_info("Device refused sync: %s", error_description);
 		goto out;
 	}
 
+	// did the device cancel the session?
 	if (!strcmp(response_type, "SDMessageCancelSession")) {
-		char *reason = NULL;
 		err = MOBILESYNC_E_CANCELLED;
-		plist_get_string_val(plist_array_get_item(msg, 2), &reason);
-		debug_info("Device cancelled: %s", reason);
-		free(reason);
-		reason = NULL;
+		plist_get_string_val(plist_array_get_item(msg, 2), error_description);
+		debug_info("Device cancelled: %s", error_description);
 		goto out;
 	}
 
@@ -313,9 +333,10 @@ mobilesync_error_t mobilesync_start(mobilesync_client_t client, const char *data
  *
  * @param client The mobilesync client
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_PLIST_ERROR if
- *    the received plist is not of valid form
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_PLIST_ERROR if the received plist is not of valid
+ * form
  */
 mobilesync_error_t mobilesync_finish(mobilesync_client_t client)
 {
@@ -409,8 +430,8 @@ static mobilesync_error_t mobilesync_get_records(mobilesync_client_t client, con
  *
  * @param client The mobilesync client
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
  */
 mobilesync_error_t mobilesync_get_all_records_from_device(mobilesync_client_t client)
 {
@@ -424,8 +445,8 @@ mobilesync_error_t mobilesync_get_all_records_from_device(mobilesync_client_t cl
  *
  * @param client The mobilesync client
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
  */
 mobilesync_error_t mobilesync_get_changes_from_device(mobilesync_client_t client)
 {
@@ -437,12 +458,13 @@ mobilesync_error_t mobilesync_get_changes_from_device(mobilesync_client_t client
  *
  * @param client The mobilesync client
  * @param entities A pointer to store the changed entity records as a PLIST_DICT
- * @param is_last_record A pointer to store a flag indiciating if this submission is the last one
+ * @param is_last_record A pointer to store a flag indicating if this submission is the last one
  * @param actions A pointer to additional flags the device is sending or NULL to ignore
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_CANCELLED if the device
- *    explicitly cancelled the session
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_CANCELLED if the device explicitly cancelled the
+ * session
  */
 mobilesync_error_t mobilesync_receive_changes(mobilesync_client_t client, plist_t *entities, uint8_t *is_last_record, plist_t *actions)
 {
@@ -511,13 +533,93 @@ mobilesync_error_t mobilesync_receive_changes(mobilesync_client_t client, plist_
 	return err;
 }
 
+/**
+ * Requests the device to delete all records of the current data class
+ *
+ * @note The operation must be called after starting synchronization.
+ *
+ * @param client The mobilesync client
+ *
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_PLIST_ERROR if the received plist is not of valid form
+ */
+mobilesync_error_t mobilesync_clear_all_records_on_device(mobilesync_client_t client)
+{
+	if (!client || !client->data_class) {
+		return MOBILESYNC_E_INVALID_ARG;
+	}
+
+	mobilesync_error_t err = MOBILESYNC_E_UNKNOWN_ERROR;
+	plist_t msg = NULL;
+	plist_t response_type_node = NULL;
+	char *response_type = NULL;
+
+	msg = plist_new_array();
+	plist_array_append_item(msg, plist_new_string("SDMessageClearAllRecordsOnDevice"));
+	plist_array_append_item(msg, plist_new_string(client->data_class));
+	plist_array_append_item(msg, plist_new_string(EMPTY_PARAMETER_STRING));
+
+	err = mobilesync_send(client, msg);
+
+	if (err != MOBILESYNC_E_SUCCESS) {
+		goto out;
+	}
+
+	plist_free(msg);
+	msg = NULL;
+
+	err = mobilesync_receive(client, &msg);
+
+	if (err != MOBILESYNC_E_SUCCESS) {
+		goto out;
+	}
+
+	response_type_node = plist_array_get_item(msg, 0);
+	if (!response_type_node) {
+		err = MOBILESYNC_E_PLIST_ERROR;
+		goto out;
+	}
+
+	plist_get_string_val(response_type_node, &response_type);
+	if (!response_type) {
+		err = MOBILESYNC_E_PLIST_ERROR;
+		goto out;
+	}
+
+	if (!strcmp(response_type, "SDMessageCancelSession")) {
+		char *reason = NULL;
+		err = MOBILESYNC_E_CANCELLED;
+		plist_get_string_val(plist_array_get_item(msg, 2), &reason);
+		debug_info("Device cancelled: %s", reason);
+		free(reason);
+		goto out;
+	}
+
+	if (strcmp(response_type, "SDMessageDeviceWillClearAllRecords")) {
+		err = MOBILESYNC_E_PLIST_ERROR;
+	}
+
+	out:
+	if (response_type) {
+		free(response_type);
+		response_type = NULL;
+	}
+	if (msg) {
+		plist_free(msg);
+		msg = NULL;
+	}
+
+	return err;
+}
+
 /**
  * Acknowledges to the device that the changes have been merged on the computer
  *
  * @param client The mobilesync client
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
  */
 mobilesync_error_t mobilesync_acknowledge_changes_from_device(mobilesync_client_t client)
 {
@@ -560,12 +662,15 @@ static plist_t create_process_changes_message(const char *data_class, plist_t en
  *
  * @param client The mobilesync client
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_PLIST_ERROR if
- *    the received plist is not of valid form, MOBILESYNC_E_WRONG_DIRECTION if the
- *    current sync direction does not permit this call, MOBILESYNC_E_CANCELLED if
- *    the device explicitly cancelled the session or MOBILESYNC_E_NOT_READY if the
- *    device is not ready to start receiving any changes
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_PLIST_ERROR if the received plist is not of valid form
+ * @retval MOBILESYNC_E_WRONG_DIRECTION if the current sync direction does
+ * not permit this call
+ * @retval MOBILESYNC_E_CANCELLED if the device explicitly cancelled the
+ * session
+ * @retval MOBILESYNC_E_NOT_READY if the device is not ready to start
+ * receiving any changes
  */
 mobilesync_error_t mobilesync_ready_to_send_changes_from_computer(mobilesync_client_t client)
 {
@@ -622,6 +727,10 @@ mobilesync_error_t mobilesync_ready_to_send_changes_from_computer(mobilesync_cli
 	err = MOBILESYNC_E_SUCCESS;
 
 	out:
+	if (response_type) {
+		free(response_type);
+		response_type = NULL;
+	}
 	if (msg) {
 		plist_free(msg);
 		msg = NULL;
@@ -631,17 +740,18 @@ mobilesync_error_t mobilesync_ready_to_send_changes_from_computer(mobilesync_cli
 }
 
 /**
- * Sends changed entitites of the currently set data class to the device
+ * Sends changed entities of the currently set data class to the device
  *
  * @param client The mobilesync client
  * @param entities The changed entity records as a PLIST_DICT
- * @param is_last_record A flag indiciating if this submission is the last one
+ * @param is_last_record A flag indicating if this submission is the last one
  * @param actions Additional actions for the device created with mobilesync_actions_new()
  *    or NULL if no actions should be passed
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_WRONG_DIRECTION if the
- *    current sync direction does not permit this call
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid,
+ * @retval MOBILESYNC_E_WRONG_DIRECTION if the current sync direction does
+ * not permit this call
  */
 mobilesync_error_t mobilesync_send_changes(mobilesync_client_t client, plist_t entities, uint8_t is_last_record, plist_t actions)
 {
@@ -677,11 +787,14 @@ mobilesync_error_t mobilesync_send_changes(mobilesync_client_t client, plist_t e
  * @param client The mobilesync client
  * @param mapping A pointer to an array plist containing a dict of identifier remappings
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid, MOBILESYNC_E_PLIST_ERROR if
- *    the received plist is not of valid form, MOBILESYNC_E_WRONG_DIRECTION if the
- *    current sync direction does not permit this call or MOBILESYNC_E_CANCELLED if
- *    the device explicitly cancelled the session
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
+ * @retval MOBILESYNC_E_PLIST_ERROR if the received plist is not of valid
+ * form
+ * @retval MOBILESYNC_E_WRONG_DIRECTION if the current sync direction does
+ * not permit this call
+ * @retval MOBILESYNC_E_CANCELLED if the device explicitly cancelled the
+ * session
  */
 mobilesync_error_t mobilesync_remap_identifiers(mobilesync_client_t client, plist_t *mapping)
 {
@@ -753,13 +866,13 @@ mobilesync_error_t mobilesync_remap_identifiers(mobilesync_client_t client, plis
 }
 
 /**
- * Cancells a running synchronization session with a device at any time.
+ * Cancels a running synchronization session with a device at any time.
  *
  * @param client The mobilesync client
  * @param reason The reason to supply to the device for cancelling
  *
- * @return MOBILESYNC_E_SUCCESS on success, MOBILESYNC_E_INVALID_ARG if
- *    one of the parameters is invalid
+ * @retval MOBILESYNC_E_SUCCESS on success
+ * @retval MOBILESYNC_E_INVALID_ARG if one of the parameters is invalid
  */
 mobilesync_error_t mobilesync_cancel(mobilesync_client_t client, const char* reason)
 {
@@ -794,7 +907,7 @@ mobilesync_error_t mobilesync_cancel(mobilesync_client_t client, const char* rea
  *   if none is known yet which for instance is true on first synchronization.
  * @param computer_anchor An arbitrary string to use as anchor for the computer.
  * 
- * @return A new mobilesync_anchors_t struct. Must be freed using mobilesync_anchors_free().
+ * @return A new #mobilesync_anchors_t struct. Must be freed using mobilesync_anchors_free().
  */
 mobilesync_anchors_t mobilesync_anchors_new(const char *device_anchor, const char *computer_anchor)
 {
@@ -816,7 +929,7 @@ mobilesync_anchors_t mobilesync_anchors_new(const char *device_anchor, const cha
 /**
  * Free memory used by anchors.
  *
- * @param anchors The anchors free.
+ * @param anchors The anchors to free.
  */
 void mobilesync_anchors_free(mobilesync_anchors_t anchors)
 {
@@ -851,7 +964,7 @@ plist_t mobilesync_actions_new()
  * @note The known keys so far are "SyncDeviceLinkEntityNamesKey" which expects
  *       an array of entity names, followed by a count paramter as well as
  *       "SyncDeviceLinkAllRecordsOfPulledEntityTypeSentKey" which expects an
- *       integer to use as a boolean value indiciating that the device should
+ *       integer to use as a boolean value indicating that the device should
  *       link submitted changes and report remapped identifiers.
  */
 void mobilesync_actions_add(plist_t actions, ...)
@@ -874,10 +987,10 @@ void mobilesync_actions_add(plist_t actions, ...)
 				plist_array_append_item(array, plist_new_string(entity_names[i]));
 			}
 
-			plist_dict_insert_item(actions, key, array);
+			plist_dict_set_item(actions, key, array);
 		} else if (!strcmp(key, "SyncDeviceLinkAllRecordsOfPulledEntityTypeSentKey")) {
 			int link_records = va_arg(args, int);
-			plist_dict_insert_item(actions, key, plist_new_bool(link_records));
+			plist_dict_set_item(actions, key, plist_new_bool(link_records));
 		}
 		free(key);
 		key = NULL;
