@@ -26,7 +26,7 @@
 
 #include "sbservices.h"
 #include "property_list_service.h"
-#include "debug.h"
+#include "common/debug.h"
 
 /**
  * Locks an sbservices client, used for thread safety.
@@ -36,7 +36,7 @@
 static void sbs_lock(sbservices_client_t client)
 {
 	debug_info("SBServices: Locked");
-	g_mutex_lock(client->mutex);
+	mutex_lock(&client->mutex);
 }
 
 /**
@@ -47,7 +47,7 @@ static void sbs_lock(sbservices_client_t client)
 static void sbs_unlock(sbservices_client_t client)
 {
 	debug_info("SBServices: Unlocked");
-	g_mutex_unlock(client->mutex);
+	mutex_unlock(&client->mutex);
 }
 
 /**
@@ -61,55 +61,68 @@ static void sbs_unlock(sbservices_client_t client)
  */
 static sbservices_error_t sbservices_error(property_list_service_error_t err)
 {
-       switch (err) {
-                case PROPERTY_LIST_SERVICE_E_SUCCESS:
-                        return SBSERVICES_E_SUCCESS;
-                case PROPERTY_LIST_SERVICE_E_INVALID_ARG:
-                        return SBSERVICES_E_INVALID_ARG;
-                case PROPERTY_LIST_SERVICE_E_PLIST_ERROR:
-                        return SBSERVICES_E_PLIST_ERROR;
-                case PROPERTY_LIST_SERVICE_E_MUX_ERROR:
-                        return SBSERVICES_E_CONN_FAILED;
-                default:
-                        break;
-        }
-        return SBSERVICES_E_UNKNOWN_ERROR;
+	switch (err) {
+		case PROPERTY_LIST_SERVICE_E_SUCCESS:
+			return SBSERVICES_E_SUCCESS;
+		case PROPERTY_LIST_SERVICE_E_INVALID_ARG:
+			return SBSERVICES_E_INVALID_ARG;
+		case PROPERTY_LIST_SERVICE_E_PLIST_ERROR:
+			return SBSERVICES_E_PLIST_ERROR;
+		case PROPERTY_LIST_SERVICE_E_MUX_ERROR:
+			return SBSERVICES_E_CONN_FAILED;
+		default:
+			break;
+	}
+	return SBSERVICES_E_UNKNOWN_ERROR;
 }
 
 /**
  * Connects to the springboardservices service on the specified device.
  *
  * @param device The device to connect to.
- * @param port Destination port (usually given by lockdownd_start_service).
+ * @param service The service descriptor returned by lockdownd_start_service.
  * @param client Pointer that will point to a newly allocated
  *     sbservices_client_t upon successful return.
  *
  * @return SBSERVICES_E_SUCCESS on success, SBSERVICES_E_INVALID_ARG when
  *     client is NULL, or an SBSERVICES_E_* error code otherwise.
  */
-sbservices_error_t sbservices_client_new(idevice_t device, uint16_t port, sbservices_client_t *client)
+sbservices_error_t sbservices_client_new(idevice_t device, lockdownd_service_descriptor_t service, sbservices_client_t *client)
 {
-	/* makes sure thread environment is available */
-	if (!g_thread_supported())
-		g_thread_init(NULL);
-
-	if (!device)
-		return SBSERVICES_E_INVALID_ARG;
-
 	property_list_service_client_t plistclient = NULL;
-	sbservices_error_t err = sbservices_error(property_list_service_client_new(device, port, &plistclient));
+	sbservices_error_t err = sbservices_error(property_list_service_client_new(device, service, &plistclient));
 	if (err != SBSERVICES_E_SUCCESS) {
 		return err;
 	}
 
 	sbservices_client_t client_loc = (sbservices_client_t) malloc(sizeof(struct sbservices_client_private));
 	client_loc->parent = plistclient;
-	client_loc->mutex = g_mutex_new();
+	mutex_init(&client_loc->mutex);
 
 	*client = client_loc;
 	return SBSERVICES_E_SUCCESS;
 }
 
+/**
+ * Starts a new sbservices service on the specified device and connects to it.
+ *
+ * @param device The device to connect to.
+ * @param client Pointer that will point to a newly allocated
+ *     sbservices_client_t upon successful return. Must be freed using
+ *     sbservices_client_free() after use.
+ * @param label The label to use for communication. Usually the program name.
+ *  Pass NULL to disable sending the label in requests to lockdownd.
+ *
+ * @return SBSERVICES_E_SUCCESS on success, or an SBSERVICES_E_* error
+ *     code otherwise.
+ */
+sbservices_error_t sbservices_client_start_service(idevice_t device, sbservices_client_t * client, const char* label)
+{
+	sbservices_error_t err = SBSERVICES_E_UNKNOWN_ERROR;
+	service_client_factory_start_service(device, SBSERVICES_SERVICE_NAME, (void**)client, label, SERVICE_CONSTRUCTOR(sbservices_client_new), &err);
+	return err;
+}
+
 /**
  * Disconnects an sbservices client from the device and frees up the
  * sbservices client data.
@@ -126,9 +139,7 @@ sbservices_error_t sbservices_client_free(sbservices_client_t client)
 
 	sbservices_error_t err = sbservices_error(property_list_service_client_free(client->parent));
 	client->parent = NULL;
-	if (client->mutex) {
-		g_mutex_free(client->mutex);
-	}
+	mutex_destroy(&client->mutex);
 	free(client);
 
 	return err;
@@ -156,9 +167,9 @@ sbservices_error_t sbservices_get_icon_state(sbservices_client_t client, plist_t
 	sbservices_error_t res = SBSERVICES_E_UNKNOWN_ERROR;
 
 	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "command", plist_new_string("getIconState"));
+	plist_dict_set_item(dict, "command", plist_new_string("getIconState"));
 	if (format_version) {
-		plist_dict_insert_item(dict, "formatVersion", plist_new_string(format_version));
+		plist_dict_set_item(dict, "formatVersion", plist_new_string(format_version));
 	}
 
 	sbs_lock(client);
@@ -205,8 +216,8 @@ sbservices_error_t sbservices_set_icon_state(sbservices_client_t client, plist_t
 	sbservices_error_t res = SBSERVICES_E_UNKNOWN_ERROR;
 
 	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "command", plist_new_string("setIconState"));
-	plist_dict_insert_item(dict, "iconState", plist_copy(newstate));
+	plist_dict_set_item(dict, "command", plist_new_string("setIconState"));
+	plist_dict_set_item(dict, "iconState", plist_copy(newstate));
 
 	sbs_lock(client);
 
@@ -246,8 +257,8 @@ sbservices_error_t sbservices_get_icon_pngdata(sbservices_client_t client, const
 	sbservices_error_t res = SBSERVICES_E_UNKNOWN_ERROR;
 
 	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "command", plist_new_string("getIconPNGData"));
-	plist_dict_insert_item(dict, "bundleId", plist_new_string(bundleId));
+	plist_dict_set_item(dict, "command", plist_new_string("getIconPNGData"));
+	plist_dict_set_item(dict, "bundleId", plist_new_string(bundleId));
 
 	sbs_lock(client);
 
@@ -273,7 +284,53 @@ leave_unlock:
 	}
 	sbs_unlock(client);
 	return res;
+}
 
+/**
+ * Gets the interface orientation of the device.
+ *
+ * @param client The connected sbservices client to use.
+ * @param interface_orientation The interface orientation upon successful return.
+ *
+ * @return SBSERVICES_E_SUCCESS on success, SBSERVICES_E_INVALID_ARG when
+ *     client or state is invalid, or an SBSERVICES_E_* error code otherwise.
+ */
+sbservices_error_t sbservices_get_interface_orientation(sbservices_client_t client, sbservices_interface_orientation_t* interface_orientation)
+{
+	if (!client || !client->parent || !interface_orientation)
+		return SBSERVICES_E_INVALID_ARG;
+
+	sbservices_error_t res = SBSERVICES_E_UNKNOWN_ERROR;
+
+	plist_t dict = plist_new_dict();
+	plist_dict_set_item(dict, "command", plist_new_string("getInterfaceOrientation"));
+
+	sbs_lock(client);
+
+	res = sbservices_error(property_list_service_send_binary_plist(client->parent, dict));
+	if (res != SBSERVICES_E_SUCCESS) {
+		debug_info("could not send plist, error %d", res);
+		goto leave_unlock;
+	}
+	plist_free(dict);
+	dict = NULL;
+
+	res = sbservices_error(property_list_service_receive_plist(client->parent, &dict));
+	if (res	== SBSERVICES_E_SUCCESS) {
+		plist_t node = plist_dict_get_item(dict, "interfaceOrientation");
+		if (node) {
+			uint64_t value = SBSERVICES_INTERFACE_ORIENTATION_UNKNOWN;
+			plist_get_uint_val(node, &value);
+			*interface_orientation = (sbservices_interface_orientation_t)value;
+		}
+	}
+
+leave_unlock:
+	if (dict) {
+		plist_free(dict);
+	}
+	sbs_unlock(client);
+	return res;
 }
 
 /**
@@ -298,7 +355,7 @@ sbservices_error_t sbservices_get_home_screen_wallpaper_pngdata(sbservices_clien
 	sbservices_error_t res = SBSERVICES_E_UNKNOWN_ERROR;
 
 	plist_t dict = plist_new_dict();
-	plist_dict_insert_item(dict, "command", plist_new_string("getHomeScreenWallpaperPNGData"));
+	plist_dict_set_item(dict, "command", plist_new_string("getHomeScreenWallpaperPNGData"));
 
 	sbs_lock(client);
 
