@@ -20,14 +20,23 @@
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA 
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <signal.h>
-#include <glib.h>
+#ifdef HAVE_OPENSSL
+#include <openssl/sha.h>
+#else
 #include <gcrypt.h>
+#endif
 #include <unistd.h>
+#include <ctype.h>
+#include <time.h>
 
 #include <libimobiledevice/libimobiledevice.h>
 #include <libimobiledevice/lockdown.h>
@@ -41,9 +50,14 @@
 #define LOCK_ATTEMPTS 50
 #define LOCK_WAIT 200000
 
+#ifdef WIN32
+#include <windows.h>
+#define sleep(x) Sleep(x*1000)
+#endif
+
 static mobilebackup_client_t mobilebackup = NULL;
 static lockdownd_client_t client = NULL;
-static idevice_t phone = NULL;
+static idevice_t device = NULL;
 
 static int quit_flag = 0;
 
@@ -66,7 +80,11 @@ enum device_link_file_status_t {
 
 static void sha1_of_data(const char *input, uint32_t size, unsigned char *hash_out)
 {
+#ifdef HAVE_OPENSSL
+	SHA1((const unsigned char*)input, size, hash_out);
+#else
 	gcry_md_hash_buffer(GCRY_MD_SHA1, hash_out, input, size);
+#endif
 }
 
 static int compare_hash(const unsigned char *hash1, const unsigned char *hash2, int hash_len)
@@ -82,6 +100,10 @@ static int compare_hash(const unsigned char *hash1, const unsigned char *hash2,
 
 static void compute_datahash(const char *path, const char *destpath, uint8_t greylist, const char *domain, const char *appid, const char *version, unsigned char *hash_out)
 {
+#ifdef HAVE_OPENSSL
+	SHA_CTX sha1;
+	SHA1_Init(&sha1);
+#else
 	gcry_md_hd_t hd = NULL;
 	gcry_md_open(&hd, GCRY_MD_SHA1, 0);
 	if (!hd) {
@@ -89,44 +111,103 @@ static void compute_datahash(const char *path, const char *destpath, uint8_t gre
 		return;
 	}
 	gcry_md_reset(hd);
-
+#endif
 	FILE *f = fopen(path, "rb");
 	if (f) {
 		unsigned char buf[16384];
 		size_t len;
 		while ((len = fread(buf, 1, 16384, f)) > 0) {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, buf, len);
+#else
 			gcry_md_write(hd, buf, len);
+#endif
 		}
 		fclose(f);
+#ifdef HAVE_OPENSSL
+		SHA1_Update(&sha1, destpath, strlen(destpath));
+		SHA1_Update(&sha1, ";", 1);
+#else
 		gcry_md_write(hd, destpath, strlen(destpath));
 		gcry_md_write(hd, ";", 1);
+#endif
 		if (greylist == 1) {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, "true", 4);
+#else
 			gcry_md_write(hd, "true", 4);
+#endif
 		} else {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, "false", 5);
+#else
 			gcry_md_write(hd, "false", 5);
+#endif
 		}
+#ifdef HAVE_OPENSSL
+		SHA1_Update(&sha1, ";", 1);
+#else
 		gcry_md_write(hd, ";", 1);
+#endif
 		if (domain) {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, domain, strlen(domain));
+#else
 			gcry_md_write(hd, domain, strlen(domain));
+#endif
 		} else {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, "(null)", 6);
+#else
 			gcry_md_write(hd, "(null)", 6);
+#endif
 		}
+#ifdef HAVE_OPENSSL
+		SHA1_Update(&sha1, ";", 1);
+#else
 		gcry_md_write(hd, ";", 1);
+#endif
 		if (appid) {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, appid, strlen(appid));
+#else
 			gcry_md_write(hd, appid, strlen(appid));
+#endif
 		} else {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, "(null)", 6);
+#else
 			gcry_md_write(hd, "(null)", 6);
+#endif
 		}
+#ifdef HAVE_OPENSSL
+		SHA1_Update(&sha1, ";", 1);
+#else
 		gcry_md_write(hd, ";", 1);
+#endif
 		if (version) {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, version, strlen(version));
+#else
 			gcry_md_write(hd, version, strlen(version));
+#endif
 		} else {
+#ifdef HAVE_OPENSSL
+			SHA1_Update(&sha1, "(null)", 6);
+#else
 			gcry_md_write(hd, "(null)", 6);
+#endif
 		}
+#ifdef HAVE_OPENSSL
+		SHA1_Final(hash_out, &sha1);
+#else
 		unsigned char *newhash = gcry_md_read(hd, GCRY_MD_SHA1);
 		memcpy(hash_out, newhash, 20);
+#endif
 	}
+#ifndef HAVE_OPENSSL
 	gcry_md_close(hd);
+#endif
 }
 
 static void print_hash(const unsigned char *hash, int len)
@@ -147,14 +228,68 @@ static void notify_cb(const char *notification, void *userdata)
 	}
 }
 
-static plist_t mobilebackup_factory_info_plist_new()
+static char *str_toupper(char* str)
+{
+	char *res = strdup(str);
+	unsigned int i;
+	for (i = 0; i < strlen(res); i++) {
+		res[i] = toupper(res[i]);
+	}
+	return res;
+}
+
+static char* build_path(const char* elem, ...)
+{
+	if (!elem) return NULL;
+	va_list args;
+	int len = strlen(elem)+1;
+	va_start(args, elem);
+	char *arg = va_arg(args, char*);
+	while (arg) {
+		len += strlen(arg)+1;
+		arg = va_arg(args, char*);
+	}
+	va_end(args);
+
+	char* out = (char*)malloc(len);
+	strcpy(out, elem);
+
+	va_start(args, elem);
+	arg = va_arg(args, char*);
+	while (arg) {
+		strcat(out, "/");
+		strcat(out, arg);
+		arg = va_arg(args, char*);
+	}
+	va_end(args);
+	return out;
+}
+
+static char* format_size_for_display(uint64_t size)
+{
+	char buf[32];
+	double sz;
+	if (size >= 1000000000LL) {
+		sz = ((double)size / 1000000000.0f);
+		sprintf(buf, "%0.1f GB", sz);
+	} else if (size >= 1000000LL) {
+		sz = ((double)size / 1000000.0f);
+		sprintf(buf, "%0.1f MB", sz);
+	} else if (size >= 1000LL) {
+		sz = ((double)size / 1000.0f);
+		sprintf(buf, "%0.1f kB", sz);
+	} else {
+		sprintf(buf, "%d Bytes", (int)size);
+	}
+	return strdup(buf);
+}
+
+static plist_t mobilebackup_factory_info_plist_new(const char* udid)
 {
 	/* gather data from lockdown */
-	GTimeVal tv = {0, 0};
 	plist_t value_node = NULL;
 	plist_t root_node = NULL;
-	char *uuid = NULL;
-	char *uuid_uppercase = NULL;
+	char *udid_uppercase = NULL;
 
 	plist_t ret = plist_new_dict();
 
@@ -163,45 +298,42 @@ static plist_t mobilebackup_factory_info_plist_new()
 
 	/* set fields we understand */
 	value_node = plist_dict_get_item(root_node, "BuildVersion");
-	plist_dict_insert_item(ret, "Build Version", plist_copy(value_node));
+	plist_dict_set_item(ret, "Build Version", plist_copy(value_node));
 
 	value_node = plist_dict_get_item(root_node, "DeviceName");
-	plist_dict_insert_item(ret, "Device Name", plist_copy(value_node));
-	plist_dict_insert_item(ret, "Display Name", plist_copy(value_node));
+	plist_dict_set_item(ret, "Device Name", plist_copy(value_node));
+	plist_dict_set_item(ret, "Display Name", plist_copy(value_node));
 
 	/* FIXME: How is the GUID generated? */
-	plist_dict_insert_item(ret, "GUID", plist_new_string("---"));
+	plist_dict_set_item(ret, "GUID", plist_new_string("---"));
 
 	value_node = plist_dict_get_item(root_node, "InternationalMobileEquipmentIdentity");
 	if (value_node)
-		plist_dict_insert_item(ret, "IMEI", plist_copy(value_node));
+		plist_dict_set_item(ret, "IMEI", plist_copy(value_node));
 
-	g_get_current_time(&tv);
-	plist_dict_insert_item(ret, "Last Backup Date", plist_new_date(tv.tv_sec, tv.tv_usec));
+	plist_dict_set_item(ret, "Last Backup Date", plist_new_date(time(NULL), 0));
 
 	value_node = plist_dict_get_item(root_node, "ProductType");
-	plist_dict_insert_item(ret, "Product Type", plist_copy(value_node));
+	plist_dict_set_item(ret, "Product Type", plist_copy(value_node));
 
 	value_node = plist_dict_get_item(root_node, "ProductVersion");
-	plist_dict_insert_item(ret, "Product Version", plist_copy(value_node));
+	plist_dict_set_item(ret, "Product Version", plist_copy(value_node));
 
 	value_node = plist_dict_get_item(root_node, "SerialNumber");
-	plist_dict_insert_item(ret, "Serial Number", plist_copy(value_node));
+	plist_dict_set_item(ret, "Serial Number", plist_copy(value_node));
 
 	value_node = plist_dict_get_item(root_node, "UniqueDeviceID");
-	idevice_get_uuid(phone, &uuid);
-	plist_dict_insert_item(ret, "Target Identifier", plist_new_string(uuid));
+	plist_dict_set_item(ret, "Target Identifier", plist_new_string(udid));
 
 	/* uppercase */
-	uuid_uppercase = g_ascii_strup(uuid, -1);
-	plist_dict_insert_item(ret, "Unique Identifier", plist_new_string(uuid_uppercase));
-	free(uuid_uppercase);
-	free(uuid);
+	udid_uppercase = str_toupper((char*)udid);
+	plist_dict_set_item(ret, "Unique Identifier", plist_new_string(udid_uppercase));
+	free(udid_uppercase);
 
 	/* FIXME: Embed files as <data> nodes */
 	plist_t files = plist_new_dict();
-	plist_dict_insert_item(ret, "iTunes Files", files);
-	plist_dict_insert_item(ret, "iTunes Version", plist_new_string("9.0.2"));
+	plist_dict_set_item(ret, "iTunes Files", files);
+	plist_dict_set_item(ret, "iTunes Version", plist_new_string("9.0.2"));
 
 	plist_free(root_node);
 
@@ -210,15 +342,13 @@ static plist_t mobilebackup_factory_info_plist_new()
 
 static void mobilebackup_info_update_last_backup_date(plist_t info_plist)
 {
-	GTimeVal tv = {0, 0};
 	plist_t node = NULL;
 
 	if (!info_plist)
 		return;
 
-	g_get_current_time(&tv);
 	node = plist_dict_get_item(info_plist, "Last Backup Date");
-	plist_set_date_val(node, tv.tv_sec, tv.tv_usec);
+	plist_set_date_val(node, time(NULL), 0);
 
 	node = NULL;
 }
@@ -321,11 +451,14 @@ static int plist_strcmp(plist_t node, const char *str)
 	return ret;
 }
 
-static gchar *mobilebackup_build_path(const char *backup_directory, const char *name, const char *extension)
+static char *mobilebackup_build_path(const char *backup_directory, const char *name, const char *extension)
 {
-	gchar *filename = g_strconcat(name, extension, NULL);
-	gchar *path = g_build_path(G_DIR_SEPARATOR_S, backup_directory, filename, NULL);
-	g_free(filename);
+	char* filename = (char*)malloc(strlen(name)+(extension == NULL ? 0: strlen(extension))+1);
+	strcpy(filename, name);
+	if (extension != NULL)
+		strcat(filename, extension);
+	char *path = build_path(backup_directory, filename, NULL);
+	free(filename);
 	return path;
 }
 
@@ -333,8 +466,8 @@ static void mobilebackup_write_status(const char *path, int status)
 {
 	struct stat st;
 	plist_t status_plist = plist_new_dict();
-	plist_dict_insert_item(status_plist, "Backup Success", plist_new_bool(status));
-	gchar *file_path = mobilebackup_build_path(path, "Status", ".plist");
+	plist_dict_set_item(status_plist, "Backup Success", plist_new_bool(status));
+	char *file_path = mobilebackup_build_path(path, "Status", ".plist");
 
 	if (stat(file_path, &st) == 0)
 		remove(file_path);
@@ -344,17 +477,17 @@ static void mobilebackup_write_status(const char *path, int status)
 	plist_free(status_plist);
 	status_plist = NULL;
 
-	g_free(file_path);
+	free(file_path);
 }
 
 static int mobilebackup_read_status(const char *path)
 {
 	int ret = -1;
 	plist_t status_plist = NULL;
-	gchar *file_path = mobilebackup_build_path(path, "Status", ".plist");
+	char *file_path = mobilebackup_build_path(path, "Status", ".plist");
 
 	plist_read_from_filename(&status_plist, file_path);
-	g_free(file_path);
+	free(file_path);
 	if (!status_plist) {
 		printf("Could not read Status.plist!\n");
 		return ret;
@@ -387,7 +520,7 @@ static int mobilebackup_info_is_current_device(plist_t info)
 	/* get basic device information in one go */
 	lockdownd_get_value(client, NULL, NULL, &root_node);
 
-	/* verify UUID */
+	/* verify UDID */
 	value_node = plist_dict_get_item(root_node, "UniqueDeviceID");
 	node = plist_dict_get_item(info, "Target Identifier");
 
@@ -435,14 +568,14 @@ static int mobilebackup_info_is_current_device(plist_t info)
 static int mobilebackup_delete_backup_file_by_hash(const char *backup_directory, const char *hash)
 {
 	int ret = 0;
-	gchar *path = mobilebackup_build_path(backup_directory, hash, ".mddata");
+	char *path = mobilebackup_build_path(backup_directory, hash, ".mddata");
 	printf("Removing \"%s\" ", path);
 	if (!remove( path ))
 		ret = 1;
 	else
 		ret = 0;
 
-	g_free(path);
+	free(path);
 
 	if (!ret)
 		return ret;
@@ -454,7 +587,7 @@ static int mobilebackup_delete_backup_file_by_hash(const char *backup_directory,
 	else
 		ret = 0;
 
-	g_free(path);
+	free(path);
 
 	return ret;
 }
@@ -591,9 +724,9 @@ static int mobilebackup_check_file_integrity(const char *backup_directory, const
 		hash_ok = 1;
 	}
 
-	g_free(domain);
-	g_free(version);
-	g_free(destpath);
+	free(domain);
+	free(version);
+	free(destpath);
 
 	if (!hash_ok) {
 		printf("\r\n");
@@ -605,25 +738,25 @@ static int mobilebackup_check_file_integrity(const char *backup_directory, const
 		printf("\n");
 		res = 0;
 	}
-	g_free(data_hash);
+	free(data_hash);
 	plist_free(mdinfo);
 	return res;
 }
 
 static void do_post_notification(const char *notification)
 {
-	uint16_t nport = 0;
+	lockdownd_service_descriptor_t service = NULL;
 	np_client_t np;
 
 	if (!client) {
-		if (lockdownd_client_new_with_handshake(phone, &client, "idevicebackup") != LOCKDOWN_E_SUCCESS) {
+		if (lockdownd_client_new_with_handshake(device, &client, "idevicebackup") != LOCKDOWN_E_SUCCESS) {
 			return;
 		}
 	}
 
-	lockdownd_start_service(client, NP_SERVICE_NAME, &nport);
-	if (nport) {
-		np_client_new(phone, nport, &np);
+	lockdownd_start_service(client, NP_SERVICE_NAME, &service);
+	if (service && service->port) {
+		np_client_new(device, service, &np);
 		if (np) {
 			np_post_notification(np, notification);
 			np_client_free(np);
@@ -631,6 +764,11 @@ static void do_post_notification(const char *notification)
 	} else {
 		printf("Could not start %s\n", NP_SERVICE_NAME);
 	}
+
+	if (service) {
+		lockdownd_service_descriptor_free(service);
+		service = NULL;
+	}
 }
 
 static void print_progress(double progress)
@@ -676,7 +814,7 @@ static void print_usage(int argc, char **argv)
 	printf("  restore\tRestores a device backup from DIRECTORY.\n\n");
 	printf("options:\n");
 	printf("  -d, --debug\t\tenable communication debugging\n");
-	printf("  -u, --uuid UUID\ttarget specific device by its 40-digit device UUID\n");
+	printf("  -u, --udid UDID\ttarget specific device by its 40-digit device UDID\n");
 	printf("  -h, --help\t\tprints usage information\n");
 	printf("\n");
 }
@@ -685,9 +823,8 @@ int main(int argc, char *argv[])
 {
 	idevice_error_t ret = IDEVICE_E_UNKNOWN_ERROR;
 	int i;
-	char uuid[41];
-	uint16_t port = 0;
-	uuid[0] = 0;
+	char* udid = NULL;
+	lockdownd_service_descriptor_t service = NULL;
 	int cmd = -1;
 	int is_full_backup = 0;
 	char *backup_directory = NULL;
@@ -705,9 +842,11 @@ int main(int argc, char *argv[])
 
 	/* we need to exit cleanly on running backups and restores or we cause havok */
 	signal(SIGINT, clean_exit);
-	signal(SIGQUIT, clean_exit);
 	signal(SIGTERM, clean_exit);
+#ifndef WIN32
+	signal(SIGQUIT, clean_exit);
 	signal(SIGPIPE, SIG_IGN);
+#endif
 
 	/* parse cmdline args */
 	for (i = 1; i < argc; i++) {
@@ -715,13 +854,13 @@ int main(int argc, char *argv[])
 			idevice_set_debug_level(1);
 			continue;
 		}
-		else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--uuid")) {
+		else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--udid")) {
 			i++;
 			if (!argv[i] || (strlen(argv[i]) != 40)) {
 				print_usage(argc, argv);
 				return 0;
 			}
-			strcpy(uuid, argv[i]);
+			udid = strdup(argv[i]);
 			continue;
 		}
 		else if (!strcmp(argv[i], "-h") || !strcmp(argv[i], "--help")) {
@@ -766,7 +905,7 @@ int main(int argc, char *argv[])
 	char *info_path = mobilebackup_build_path(backup_directory, "Info", ".plist");
 	if (cmd == CMD_RESTORE) {
 		if (stat(info_path, &st) != 0) {
-			g_free(info_path);
+			free(info_path);
 			printf("ERROR: Backup directory \"%s\" is invalid. No Info.plist found.\n", backup_directory);
 			return -1;
 		}
@@ -774,32 +913,32 @@ int main(int argc, char *argv[])
 
 	printf("Backup directory is \"%s\"\n", backup_directory);
 
-	if (uuid[0] != 0) {
-		ret = idevice_new(&phone, uuid);
+	if (udid) {
+		ret = idevice_new(&device, udid);
 		if (ret != IDEVICE_E_SUCCESS) {
-			printf("No device found with uuid %s, is it plugged in?\n", uuid);
+			printf("No device found with udid %s, is it plugged in?\n", udid);
 			return -1;
 		}
 	}
 	else
 	{
-		ret = idevice_new(&phone, NULL);
+		ret = idevice_new(&device, NULL);
 		if (ret != IDEVICE_E_SUCCESS) {
 			printf("No device found, is it plugged in?\n");
 			return -1;
 		}
 	}
 
-	if (LOCKDOWN_E_SUCCESS != lockdownd_client_new_with_handshake(phone, &client, "idevicebackup")) {
-		idevice_free(phone);
+	if (LOCKDOWN_E_SUCCESS != lockdownd_client_new_with_handshake(device, &client, "idevicebackup")) {
+		idevice_free(device);
 		return -1;
 	}
 
 	/* start notification_proxy */
 	np_client_t np = NULL;
-	ret = lockdownd_start_service(client, NP_SERVICE_NAME, &port);
-	if ((ret == LOCKDOWN_E_SUCCESS) && port) {
-		np_client_new(phone, port, &np);
+	ret = lockdownd_start_service(client, NP_SERVICE_NAME, &service);
+	if ((ret == LOCKDOWN_E_SUCCESS) && service && service->port) {
+		np_client_new(device, service, &np);
 		np_set_notify_callback(np, notify_cb, NULL);
 		const char *noties[5] = {
 			NP_SYNC_CANCEL_REQUEST,
@@ -816,19 +955,29 @@ int main(int argc, char *argv[])
 	afc_client_t afc = NULL;
 	if (cmd == CMD_BACKUP) {
 		/* start AFC, we need this for the lock file */
-		port = 0;
-		ret = lockdownd_start_service(client, "com.apple.afc", &port);
-		if ((ret == LOCKDOWN_E_SUCCESS) && port) {
-			afc_client_new(phone, port, &afc);
+		service->port = 0;
+		service->ssl_enabled = 0;
+		ret = lockdownd_start_service(client, "com.apple.afc", &service);
+		if ((ret == LOCKDOWN_E_SUCCESS) && service->port) {
+			afc_client_new(device, service, &afc);
 		}
 	}
 
+	if (service) {
+		lockdownd_service_descriptor_free(service);
+		service = NULL;
+	}
+
 	/* start mobilebackup service and retrieve port */
-	port = 0;
-	ret = lockdownd_start_service(client, MOBILEBACKUP_SERVICE_NAME, &port);
-	if ((ret == LOCKDOWN_E_SUCCESS) && port) {
-		printf("Started \"%s\" service on port %d.\n", MOBILEBACKUP_SERVICE_NAME, port);
-		mobilebackup_client_new(phone, port, &mobilebackup);
+	ret = lockdownd_start_service(client, MOBILEBACKUP_SERVICE_NAME, &service);
+	if ((ret == LOCKDOWN_E_SUCCESS) && service && service->port) {
+		printf("Started \"%s\" service on port %d.\n", MOBILEBACKUP_SERVICE_NAME, service->port);
+		mobilebackup_client_new(device, service, &mobilebackup);
+
+		if (service) {
+			lockdownd_service_descriptor_free(service);
+			service = NULL;
+		}
 
 		/* check abort conditions */
 		if (quit_flag > 0) {
@@ -932,10 +1081,10 @@ int main(int argc, char *argv[])
 				}
 				remove(info_path);
 				printf("Creating Info.plist for new backup.\n");
-				info_plist = mobilebackup_factory_info_plist_new();
+				info_plist = mobilebackup_factory_info_plist_new(udid);
 				plist_write_to_filename(info_plist, info_path, PLIST_FORMAT_XML);
 			}
-			g_free(info_path);
+			free(info_path);
 
 			plist_free(info_plist);
 			info_plist = NULL;
@@ -985,7 +1134,7 @@ int main(int argc, char *argv[])
 			char *filename_mddata = NULL;
 			char *filename_source = NULL;
 			char *format_size = NULL;
-			gboolean is_manifest = FALSE;
+			int is_manifest = 0;
 			uint8_t b = 0;
 
 			/* process series of DLSendFile messages */
@@ -1010,9 +1159,9 @@ int main(int argc, char *argv[])
 					node = plist_dict_get_item(node_tmp, "BackupTotalSizeKey");
 					if (node) {
 						plist_get_uint_val(node, &backup_total_size);
-						format_size = g_format_size_for_display(backup_total_size);
+						format_size = format_size_for_display(backup_total_size);
 						printf("Backup data requires %s on the disk.\n", format_size);
-						g_free(format_size);
+						free(format_size);
 					}
 				}
 
@@ -1027,7 +1176,7 @@ int main(int argc, char *argv[])
 				if (node) {
 					plist_get_bool_val(node, &b);
 				}
-				is_manifest = (b == 1) ? TRUE: FALSE;
+				is_manifest = (b == 1) ? 1 : 0;
 
 				if ((hunk_index == 0) && (!is_manifest)) {
 					/* get source filename */
@@ -1040,17 +1189,17 @@ int main(int argc, char *argv[])
 					plist_get_uint_val(node, &file_size);
 					backup_real_size += file_size;
 
-					format_size = g_format_size_for_display(backup_real_size);
+					format_size = format_size_for_display(backup_real_size);
 					printf("(%s", format_size);
-					g_free(format_size);
+					free(format_size);
 
-					format_size = g_format_size_for_display(backup_total_size);
+					format_size = format_size_for_display(backup_total_size);
 					printf("/%s): ", format_size);
-					g_free(format_size);
+					free(format_size);
 
-					format_size = g_format_size_for_display(file_size);
+					format_size = format_size_for_display(file_size);
 					printf("Receiving file %s (%s)... \n", filename_source, format_size);
-					g_free(format_size);
+					free(format_size);
 
 					if (filename_source)
 						free(filename_source);
@@ -1073,7 +1222,7 @@ int main(int argc, char *argv[])
 						node = plist_dict_get_item(node_tmp, "BackupFileInfo");
 						plist_write_to_filename(node, filename_mdinfo, PLIST_FORMAT_BINARY);
 
-						g_free(filename_mdinfo);
+						free(filename_mdinfo);
 					}
 
 					file_index++;
@@ -1107,7 +1256,7 @@ int main(int argc, char *argv[])
 					free(buffer);
 					buffer = NULL;
 
-					g_free(filename_mddata);
+					free(filename_mddata);
 				}
 
 				if ((!is_manifest)) {
@@ -1145,7 +1294,7 @@ files_out:
 
 					/* remove any atomic Manifest.plist.tmp */
 					if (manifest_path)
-						g_free(manifest_path);
+						free(manifest_path);
 
 					manifest_path = mobilebackup_build_path(backup_directory, "Manifest", ".plist.tmp");
 					if (stat(manifest_path, &st) == 0)
@@ -1255,12 +1404,12 @@ files_out:
 					} else {
 						printf("Could not get AuthSignature from manifest!\n");
 					}
-					g_free(auth_sig);
+					free(auth_sig);
 				} else if (auth_ver) {
 					printf("Unknown AuthVersion '%s', cannot verify AuthSignature\n", auth_ver); 
 				}
 				plist_from_bin(bin, (uint32_t)binsize, &backup_data);
-				g_free(bin);
+				free(bin);
 			}
 			if (!backup_data) {
 				printf("Could not read plist from Manifest.plist Data key!\n");
@@ -1360,12 +1509,12 @@ files_out:
 						printf("Restoring file %s %d/%d (%d%%)... ", file_path, cur_file, total_files, (cur_file*100/total_files));
 
 						/* add additional device link file information keys */
-						plist_dict_insert_item(file_info, "DLFileAttributesKey", plist_copy(node));
-						plist_dict_insert_item(file_info, "DLFileSource", plist_new_string(file_info_path));
-						plist_dict_insert_item(file_info, "DLFileDest", plist_new_string("/tmp/RestoreFile.plist"));
-						plist_dict_insert_item(file_info, "DLFileIsEncrypted", plist_new_bool(is_encrypted));
-						plist_dict_insert_item(file_info, "DLFileOffsetKey", plist_new_uint(file_offset));
-						plist_dict_insert_item(file_info, "DLFileStatusKey", plist_new_uint(file_status));
+						plist_dict_set_item(file_info, "DLFileAttributesKey", plist_copy(node));
+						plist_dict_set_item(file_info, "DLFileSource", plist_new_string(file_info_path));
+						plist_dict_set_item(file_info, "DLFileDest", plist_new_string("/tmp/RestoreFile.plist"));
+						plist_dict_set_item(file_info, "DLFileIsEncrypted", plist_new_bool(is_encrypted));
+						plist_dict_set_item(file_info, "DLFileOffsetKey", plist_new_uint(file_offset));
+						plist_dict_set_item(file_info, "DLFileStatusKey", plist_new_uint(file_status));
 
 						/* read data from file */
 						free(file_info_path);
@@ -1382,10 +1531,10 @@ files_out:
 								file_status = DEVICE_LINK_FILE_STATUS_HUNK;
 							
 							plist_dict_remove_item(file_info, "DLFileOffsetKey");
-							plist_dict_insert_item(file_info, "DLFileOffsetKey", plist_new_uint(file_offset));
+							plist_dict_set_item(file_info, "DLFileOffsetKey", plist_new_uint(file_offset));
 
 							plist_dict_remove_item(file_info, "DLFileStatusKey");
-							plist_dict_insert_item(file_info, "DLFileStatusKey", plist_new_uint(file_status));
+							plist_dict_set_item(file_info, "DLFileStatusKey", plist_new_uint(file_status));
 
 							send_file_node = plist_new_array();
 
@@ -1466,8 +1615,8 @@ files_out:
 						tmp_node = plist_dict_get_item(node, "AppInfo");
 
 						dict = plist_new_dict();
-						plist_dict_insert_item(dict, "AppInfo", plist_copy(tmp_node));
-						plist_dict_insert_item(dict, "BackupMessageTypeKey", plist_new_string("BackupMessageRestoreApplicationSent"));
+						plist_dict_set_item(dict, "AppInfo", plist_copy(tmp_node));
+						plist_dict_set_item(dict, "BackupMessageTypeKey", plist_new_string("BackupMessageRestoreApplicationSent"));
 
 						array = plist_new_array();
 						plist_array_append_item(array, plist_new_string("DLMessageProcessMessage"));
@@ -1540,7 +1689,7 @@ files_out:
 			do_post_notification(NP_SYNC_DID_FINISH);
 		}
 		if (manifest_path)
-			g_free(manifest_path);
+			free(manifest_path);
 	} else {
 		printf("ERROR: Could not start service %s.\n", MOBILEBACKUP_SERVICE_NAME);
 		lockdownd_client_free(client);
@@ -1561,7 +1710,11 @@ files_out:
 	if (mobilebackup)
 		mobilebackup_client_free(mobilebackup);
 
-	idevice_free(phone);
+	idevice_free(device);
+
+	if (udid) {
+		free(udid);
+	}
 
 	return 0;
 }
