@@ -23,7 +23,9 @@
 #include <string.h>
 #include <errno.h>
 #include <stdlib.h>
-#include <glib.h>
+#include <time.h>
+#include <sys/time.h>
+#include <inttypes.h>
 
 #include <libimobiledevice/libimobiledevice.h>
 #include <libimobiledevice/lockdown.h>
@@ -33,9 +35,17 @@
 
 static const char *domains[] = {
 	"com.apple.disk_usage",
+	"com.apple.disk_usage.factory",
 	"com.apple.mobile.battery",
 /* FIXME: For some reason lockdownd segfaults on this, works sometimes though 
 	"com.apple.mobile.debug",. */
+	"com.apple.iqagent",
+	"com.apple.purplebuddy",
+	"com.apple.PurpleBuddy",
+	"com.apple.mobile.chaperone",
+	"com.apple.mobile.third_party_termination",
+	"com.apple.mobile.lockdownd",
+	"com.apple.mobile.lockdown_cache",
 	"com.apple.xcode.developerdomain",
 	"com.apple.international",
 	"com.apple.mobile.data_sync",
@@ -49,6 +59,8 @@ static const char *domains[] = {
 	"com.apple.mobile.software_behavior",
 	"com.apple.mobile.iTunes.SQLMusicLibraryPostProcessCommands",
 	"com.apple.mobile.iTunes.accessories",
+	"com.apple.mobile.internal", /**< iOS 4.0+ */
+	"com.apple.mobile.wireless_lockdown", /**< iOS 4.0+ */
 	"com.apple.fairplay",
 	"com.apple.iTunes",
 	"com.apple.mobile.iTunes.store",
@@ -56,6 +68,36 @@ static const char *domains[] = {
 	NULL
 };
 
+static const char base64_str[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+static const char base64_pad = '=';
+
+static char *base64encode(const unsigned char *buf, size_t size)
+{
+	if (!buf || !(size > 0)) return NULL;
+	int outlen = (size / 3) * 4;
+	char *outbuf = (char*)malloc(outlen+5); // 4 spare bytes + 1 for '\0'
+	size_t n = 0;
+	size_t m = 0;
+	unsigned char input[3];
+	unsigned int output[4];
+	while (n < size) {
+		input[0] = buf[n];
+		input[1] = (n+1 < size) ? buf[n+1] : 0;
+		input[2] = (n+2 < size) ? buf[n+2] : 0;
+		output[0] = input[0] >> 2;
+		output[1] = ((input[0] & 3) << 4) + (input[1] >> 4);
+		output[2] = ((input[1] & 15) << 2) + (input[2] >> 6);
+		output[3] = input[2] & 63;
+		outbuf[m++] = base64_str[(int)output[0]];
+		outbuf[m++] = base64_str[(int)output[1]];
+		outbuf[m++] = (n+1 < size) ? base64_str[(int)output[2]] : base64_pad;
+		outbuf[m++] = (n+2 < size) ? base64_str[(int)output[3]] : base64_pad;
+		n+=3;
+	}
+	outbuf[m] = 0; // 0-termination!
+	return outbuf;
+}
+
 static int indent_level = 0;
 
 static int is_domain_known(char *domain)
@@ -119,7 +161,7 @@ static void plist_node_to_string(plist_t node)
 	double d;
 	uint8_t b;
 	uint64_t u = 0;
-	GTimeVal tv = { 0, 0 };
+	struct timeval tv = { 0, 0 };
 
 	plist_type t;
 
@@ -136,7 +178,7 @@ static void plist_node_to_string(plist_t node)
 
 	case PLIST_UINT:
 		plist_get_uint_val(node, &u);
-		printf("%llu\n", (long long)u);
+		printf("%"PRIu64"\n", (long long)u);
 		break;
 
 	case PLIST_REAL:
@@ -158,17 +200,40 @@ static void plist_node_to_string(plist_t node)
 
 	case PLIST_DATA:
 		plist_get_data_val(node, &data, &u);
-		s = g_base64_encode((guchar *)data, u);
-		free(data);
-		printf("%s\n", s);
-		g_free(s);
+		if (u > 0) {
+			s = base64encode((unsigned char*)data, u);
+			free(data);
+			if (s) {
+				printf("%s\n", s);
+				free(s);
+			} else {
+				printf("\n");
+			}
+		} else {
+			printf("\n");
+		}
 		break;
 
 	case PLIST_DATE:
 		plist_get_date_val(node, (int32_t*)&tv.tv_sec, (int32_t*)&tv.tv_usec);
-		s = g_time_val_to_iso8601(&tv);
-		printf("%s\n", s);
-		free(s);
+		{
+			time_t ti = (time_t)tv.tv_sec;
+			struct tm *btime = localtime(&ti);
+			if (btime) {
+				s = (char*)malloc(24);
+ 				memset(s, 0, 24);
+				if (strftime(s, 24, "%Y-%m-%dT%H:%M:%SZ", btime) <= 0) {
+					free (s);
+					s = NULL;
+				}
+			}
+		}
+		if (s) {
+			printf("%s\n", s);
+			free(s);
+		} else {
+			printf("\n");
+		}
 		break;
 
 	case PLIST_ARRAY:
@@ -197,10 +262,10 @@ static void print_usage(int argc, char **argv)
 	
 	name = strrchr(argv[0], '/');
 	printf("Usage: %s [OPTIONS]\n", (name ? name + 1: argv[0]));
-	printf("Show information about a connected iPhone/iPod Touch.\n\n");
+	printf("Show information about a connected device.\n\n");
 	printf("  -d, --debug\t\tenable communication debugging\n");
 	printf("  -s, --simple\t\tuse a simple connection to avoid auto-pairing with the device\n");
-	printf("  -u, --uuid UUID\ttarget specific device by its 40-digit device UUID\n");
+	printf("  -u, --udid UDID\ttarget specific device by its 40-digit device UDID\n");
 	printf("  -q, --domain NAME\tset domain of query to NAME. Default: None\n");
 	printf("  -k, --key NAME\tonly query key specified by NAME. Default: All keys.\n");
 	printf("  -x, --xml\t\toutput information as xml plist instead of key/value pairs\n");
@@ -216,19 +281,18 @@ static void print_usage(int argc, char **argv)
 int main(int argc, char *argv[])
 {
 	lockdownd_client_t client = NULL;
-	idevice_t phone = NULL;
+	idevice_t device = NULL;
 	idevice_error_t ret = IDEVICE_E_UNKNOWN_ERROR;
 	int i;
 	int simple = 0;
 	int format = FORMAT_KEY_VALUE;
-	char uuid[41];
+	const char* udid = NULL;
 	char *domain = NULL;
 	char *key = NULL;
 	char *xml_doc = NULL;
 	uint32_t xml_length;
 	plist_t node = NULL;
 	plist_type node_type;
-	uuid[0] = 0;
 
 	/* parse cmdline args */
 	for (i = 1; i < argc; i++) {
@@ -236,13 +300,13 @@ int main(int argc, char *argv[])
 			idevice_set_debug_level(1);
 			continue;
 		}
-		else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--uuid")) {
+		else if (!strcmp(argv[i], "-u") || !strcmp(argv[i], "--udid")) {
 			i++;
 			if (!argv[i] || (strlen(argv[i]) != 40)) {
 				print_usage(argc, argv);
 				return 0;
 			}
-			strcpy(uuid, argv[i]);
+			udid = argv[i];
 			continue;
 		}
 		else if (!strcmp(argv[i], "-q") || !strcmp(argv[i], "--domain")) {
@@ -284,26 +348,20 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	if (uuid[0] != 0) {
-		ret = idevice_new(&phone, uuid);
-		if (ret != IDEVICE_E_SUCCESS) {
-			printf("No device found with uuid %s, is it plugged in?\n", uuid);
-			return -1;
-		}
-	}
-	else
-	{
-		ret = idevice_new(&phone, NULL);
-		if (ret != IDEVICE_E_SUCCESS) {
+	ret = idevice_new(&device, udid);
+	if (ret != IDEVICE_E_SUCCESS) {
+		if (udid) {
+			printf("No device found with udid %s, is it plugged in?\n", udid);
+		} else {
 			printf("No device found, is it plugged in?\n");
-			return -1;
 		}
+		return -1;
 	}
 
 	if (LOCKDOWN_E_SUCCESS != (simple ?
-			lockdownd_client_new(phone, &client, "ideviceinfo"):
-			lockdownd_client_new_with_handshake(phone, &client, "ideviceinfo"))) {
-		idevice_free(phone);
+			lockdownd_client_new(device, &client, "ideviceinfo"):
+			lockdownd_client_new_with_handshake(device, &client, "ideviceinfo"))) {
+		idevice_free(device);
 		return -1;
 	}
 
@@ -320,6 +378,7 @@ int main(int argc, char *argv[])
 				node_type = plist_get_node_type(node);
 				if (node_type == PLIST_DICT) {
 					plist_dict_to_string(node);
+					break;
 				} else if (node_type == PLIST_ARRAY) {
 					plist_array_to_string(node);
 					break;
@@ -337,7 +396,7 @@ int main(int argc, char *argv[])
 	if (domain != NULL)
 		free(domain);
 	lockdownd_client_free(client);
-	idevice_free(phone);
+	idevice_free(device);
 
 	return 0;
 }
